# Standard Library
from unittest.mock import patch

# Django
from django.test import override_settings

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Skillfarm
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.esi_stub import esi_client_stub_openapi
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import (
    create_skillfarm_character,
)

MODULE_PATH = "skillfarm.managers.characterskill"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".EveType.objects.bulk_get_or_create_esi", spec=True)
class TestCharacterSkillManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.audit = create_skillfarm_character(1001)

    def test_update_skills(self, _, mock_esi):
        # given
        mock_esi.client = esi_client_stub_openapi
        self.audit.update_skills(force_refresh=False)

        self.assertSetEqual(
            set(
                self.audit.skillfarm_skills.all().values_list("eve_type__id", flat=True)
            ),
            {1, 2},
        )
        obj = self.audit.skillfarm_skills.get(eve_type__id=1)
        self.assertEqual(obj.active_skill_level, 4)
        self.assertEqual(obj.skillpoints_in_skill, 128000)
        self.assertEqual(obj.trained_skill_level, 5)

        obj = self.audit.skillfarm_skills.get(eve_type__id=2)
        self.assertEqual(obj.active_skill_level, 2)
        self.assertEqual(obj.skillpoints_in_skill, 4000)
        self.assertEqual(obj.trained_skill_level, 4)



# Standard Library
from typing import Any

# Third Party
from ninja import NinjaAPI, Schema

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.shortcuts import get_object_or_404, render
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import UserProfile
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Skillfarm
from skillfarm import __title__
from skillfarm.api.helpers.core import (
    arabic_number_to_roman,
    generate_delete_character_button,
    generate_edit_skillsetup_button,
    generate_progressbar_html,
    generate_skillinfo_button,
    generate_status_icon_html,
    generate_toggle_notification_button,
    get_alts_queryset,
    get_character,
    get_main_character,
)
from skillfarm.api.helpers.skilldetails import (
    _calculate_sum_progress_bar,
    calculate_single_progress_bar,
)
from skillfarm.api.schema import CharacterSchema
from skillfarm.helpers import lazy
from skillfarm.models.skillfarmaudit import (
    CharacterSkill,
    CharacterSkillqueueEntry,
    SkillFarmAudit,
    SkillFarmSetup,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class OverviewSchema(Schema):
    portrait: str
    character: CharacterSchema
    action: str


class OverviewResponse(Schema):
    characters: list[OverviewSchema] | None = None


class SkillFarmDetailSchema(Schema):
    update_status: str
    notification: bool | None
    last_update: str | None
    is_extraction_ready: str
    is_filter: str
    progress: str | None = None


class SkillFarmDetailsSchema(Schema):
    character: CharacterSchema
    details: SkillFarmDetailSchema
    actions: str


class SkillFarmDetailsResponse(Schema):
    active_characters: list[SkillFarmDetailsSchema] | None = None
    inactive_characters: list[SkillFarmDetailsSchema] | None = None


class SkillFarmSetupSchema(Schema):
    character_id: int
    character_name: str
    skillset: list | None = None


class SkillFarmSetupResponse(Schema):
    setup: SkillFarmSetupSchema | None = None


class SkillFarmQueueSchema(Schema):
    skill: str
    start_sp: int
    end_sp: int
    trained_sp: int
    start_date: str | None = None
    finish_date: str | None = None
    progress: dict[str, Any] | None = None


class SkillFarmSkillSchema(Schema):
    skill: str
    level: int
    skillpoints: int


class SkillFarmInfoResponse(Schema):
    title: str
    character: CharacterSchema
    skillqueue: list[SkillFarmQueueSchema] | None = None
    skillqueue_filtered: list[SkillFarmQueueSchema] | None = None
    skills: list[SkillFarmSkillSchema] | None = None


def get_skillqueue_data(skill: CharacterSkillqueueEntry):
    """Get skillqueue data for a single skill"""
    level = arabic_number_to_roman(skill.finished_level)

    if skill.start_date is None:
        progress = 0
    else:
        progress = calculate_single_progress_bar(skill)

    if skill.start_date is None:
        start_date = "-"
    else:
        start_date = skill.start_date.strftime("%Y-%m-%d %H:%M")

    if skill.finish_date is None:
        end_date = "-"
    else:
        end_date = skill.finish_date.strftime("%Y-%m-%d %H:%M")

    skillqueue_response = SkillFarmQueueSchema(
        skill=f"{skill.eve_type.name} {level}",
        start_sp=skill.level_start_sp,
        end_sp=skill.level_end_sp,
        trained_sp=skill.training_start_sp,
        start_date=start_date,
        finish_date=end_date,
        progress={"html": generate_progressbar_html(progress), "value": progress},
    )
    return skillqueue_response


def get_filtered_skills_data(skill: CharacterSkill) -> list[dict[str, Any]] | None:
    """Get all Skills for the current character"""
    if skill.active_skill_level == 0:
        return None

    skills_response = SkillFarmSkillSchema(
        skill=f"{skill.eve_type.name} {arabic_number_to_roman(skill.active_skill_level)}",
        level=skill.active_skill_level,
        skillpoints=skill.skillpoints_in_skill,
    )
    return skills_response


class SkillFarmApiEndpoints:
    tags = ["SkillFarm"]

    # pylint: disable=too-many-statements, too-many-locals
    def __init__(self, api: NinjaAPI):

        @api.get(
            "/overview/", response={200: OverviewResponse, 403: dict}, tags=self.tags
        )
        def get_character_overview(
            request: WSGIRequest,
        ) -> OverviewResponse | tuple[int, dict]:
            """Get Character SkillFarm Overview"""
            logger.info(f"User {request.user} requested SkillFarm overview.")
            # Get visible characters
            chars_visible = SkillFarmAudit.objects.visible_eve_characters(request.user)

            # Check permissions
            if chars_visible is None:
                logger.warning(
                    f"User {request.user} tried to access SkillFarm overview without permissions."
                )
                return 403, {"error": "Permission Denied"}

            # Get Character IDs from visible Users
            chars_ids = chars_visible.values_list("character_id", flat=True)
            characters = UserProfile.objects.filter(
                main_character__isnull=False, main_character__character_id__in=chars_ids
            ).select_related("main_character")

            response_characters: list[OverviewSchema] = []
            for character in characters:
                try:
                    template = "skillfarm/partials/buttons/view.html"
                    button = format_html(
                        render_to_string(
                            template_name=template,
                            context={
                                "url": reverse(
                                    "skillfarm:skillfarm",
                                    kwargs={
                                        "character_id": character.main_character.character_id
                                    },
                                )
                            },
                        )
                    )

                    portrait = lazy.get_character_portrait_url(
                        character_id=character.main_character.character_id,
                        character_name=character.main_character.character_name,
                        as_html=True,
                    )

                    response_characters.append(
                        OverviewSchema(
                            portrait=portrait,
                            character=CharacterSchema(
                                character_id=character.main_character.character_id,
                                character_name=character.main_character.character_name,
                                corporation_id=character.main_character.corporation_id,
                                corporation_name=character.main_character.corporation_name,
                            ),
                            action=button,
                        )
                    )
                except AttributeError:
                    continue

            logger.info(
                f"User {request.user} successfully retrieved SkillFarm overview with {len(response_characters)} characters."
            )
            return OverviewResponse(characters=response_characters)

        @api.get(
            "{character_id}/details/",
            response={200: SkillFarmDetailsResponse, 403: dict},
            tags=self.tags,
        )
        def get_details(
            request: WSGIRequest, character_id: int
        ) -> SkillFarmDetailsResponse | tuple[int, dict]:
            """Get Character SkillFarm Details"""
            logger.info(
                f"User {request.user} requested SkillFarm details for character ID {character_id}."
            )
            # Get Main Character and check permissions
            perm, main = get_main_character(request, character_id)

            # Check permissions
            if perm is False:
                logger.warning(
                    f"User {request.user} tried to access SkillFarm details for character ID {character_id} without permissions."
                )
                return 403, {"error": "Permission Denied"}

            # Get all alts for the main character (including the main itself)
            characters = get_alts_queryset(main)
            skillfarm_characters = (
                SkillFarmAudit.objects.filter(character__in=characters)
                .select_related("character")
                .prefetch_related("skillfarm_skills", "skillfarm_skillqueue")
            )

            active_characters: list[SkillFarmDetailsSchema] = []
            inactive_characters: list[SkillFarmDetailsSchema] = []

            for character in skillfarm_characters:
                char_portrait = lazy.get_character_portrait_url(
                    character_id=character.character.character_id,
                    character_name=character.character.character_name,
                    as_html=True,
                )

                char = f"{char_portrait} {character.character.character_name} {character.get_status.bootstrap_icon()} - {character.notification_icon}"

                # Create the skillinfo button
                skill_info_html = generate_skillinfo_button(character=character)

                # Create the skillfarm action buttons
                actions = []
                actions.append(generate_toggle_notification_button(character=character))
                actions.append(generate_edit_skillsetup_button(character=character))
                actions.append(generate_delete_character_button(character=character))
                actions_html = format_html(
                    f'<div class="d-flex justify-content-end">{format_html("".join(actions))}</div>'
                )

                skillfarm_details = SkillFarmDetailsSchema(
                    character=CharacterSchema(
                        character_html=char,
                        character_id=character.character.character_id,
                        character_name=character.character.character_name,
                    ),
                    details=SkillFarmDetailSchema(
                        update_status=character.get_status,
                        notification=character.notification,
                        last_update=str(character.last_update),
                        is_extraction_ready=f"{skill_info_html} {character.extraction_icon}",
                        is_filter=generate_status_icon_html(character=character),
                    ),
                    actions=actions_html,
                )

                # Generate the progress bar for the skill queue
                if character.skillfarm_skillqueue.skill_in_training().exists() is False:
                    skillfarm_details.details.progress = str(_("No Active Training"))
                    inactive_characters.append(skillfarm_details)
                else:
                    skillqueue_response: list[SkillFarmQueueSchema] = []
                    # Get skillqueue data for each skill
                    for (
                        skill
                    ) in (
                        character.get_skillqueue
                    ):  # retrieve all skillqueue entries from character
                        skillqueue_response.append(get_skillqueue_data(skill))

                    # Calculate sum progress bar
                    skillfarm_details.details.progress = _calculate_sum_progress_bar(
                        skill_queue_response=skillqueue_response
                    )
                    active_characters.append(skillfarm_details)

            logger.info(
                f"User {request.user} successfully retrieved SkillFarm details for character ID {character_id}."
            )
            return SkillFarmDetailsResponse(
                active_characters=active_characters,
                inactive_characters=inactive_characters,
            )

        @api.get(
            "{character_id}/skillsetup/",
            response={200: SkillFarmSetupResponse, 403: dict, 404: dict},
            tags=self.tags,
        )
        def get_skillsetup(
            request, character_id: int
        ) -> SkillFarmSetupResponse | tuple[int, dict]:
            """Get Character SkillSet"""
            logger.info(
                f"User {request.user} requested SkillFarm skill setup for character ID {character_id}."
            )
            # Get Main Character and check permissions
            perm, character = get_character(request, character_id)

            if perm is False:
                logger.warning(
                    f"User {request.user} tried to access SkillFarm skill setup for character ID {character_id} without permissions."
                )
                return 403, {"error": "Permission Denied"}

            # Get SkillFarm Setup or 404
            skillfilter = get_object_or_404(SkillFarmSetup, character=character)
            skillset = skillfilter.skillset

            skillfarm_setup = SkillFarmSetupSchema(
                character_id=character.character.character_id,
                character_name=character.character.character_name,
                skillset=skillset,
            )

            logger.info(
                f"User {request.user} successfully retrieved SkillFarm skill setup for character ID {character_id}."
            )
            return SkillFarmSetupResponse(setup=skillfarm_setup)

        @api.get(
            "{character_id}/skillinfo/",
            response={200: dict, 403: str},
            tags=self.tags,
        )
        def get_skillinfo_details(request, character_id: int):
            """Get Character Skills and SkillQueue"""
            logger.info(
                f"User {request.user} requested SkillFarm skill info for character ID {character_id}."
            )
            # Get Main Character and check permissions
            perm, character = get_character(request, character_id)

            if perm is False:
                logger.warning(
                    f"User {request.user} tried to access SkillFarm skill info for character ID {character_id} without permissions."
                )
                return 403, "Permission Denied"

            response_skillqueue: list[SkillFarmQueueSchema] = []
            response_skillqueue_filtered: list[SkillFarmQueueSchema] = []
            # retrieve all skillqueue entries from character
            for skill in character.get_skillqueue:
                # Get skillqueue data for each skill
                skillqueue_response = get_skillqueue_data(skill)
                # Check if skill is filtered
                if character.is_filtered:
                    response_skillqueue_filtered.append(skillqueue_response)
                else:
                    response_skillqueue.append(skillqueue_response)

            response_skills: list[SkillFarmSkillSchema] = []
            if character.is_filtered and character.get_skillsetup is not None:
                # retrieve all skill entries from character
                for skill in character.get_skills:
                    if skill.eve_type.name in character.get_skillsetup.skillset:
                        # Get skill data for each skill
                        skill_data = get_filtered_skills_data(skill)
                        if skill_data is not None:
                            response_skills.append(skill_data)

            skillinfo_response = SkillFarmInfoResponse(
                title=str(_("Skill Info")),
                character=CharacterSchema(
                    character_id=character.character.character_id,
                    character_name=character.character.character_name,
                ),
                skillqueue=response_skillqueue,
                skillqueue_filtered=response_skillqueue_filtered,
                skills=response_skills,
            )
            logger.info(
                f"User {request.user} successfully retrieved SkillFarm skill info for character ID {character_id}."
            )
            return render(
                request,
                "skillfarm/partials/modals/view_skillqueue.html",
                context=skillinfo_response.__dict__,
            )



# Django
from django.test import TestCase
from django.utils import timezone

# AA Skillfarm
from skillfarm.models.skillfarmaudit import (
    CharacterSkill,
    CharacterSkillqueueEntry,
    SkillFarmAudit,
    SkillFarmSetup,
)
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import (
    create_skillfarm_character,
    create_update_status,
)

MODULE_PATH = "skillfarm.models.skillfarmaudit"


class TestSkillfarmModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = create_skillfarm_character(1001)

    def test_should_return_string_audit(self):
        """Test should return the Audit Character Data"""
        self.assertEqual(str(self.audit), "Gneuten - Active: True - Status: incomplete")

    def test_should_return_esi_scopes(self):
        """Test should return the ESI Scopes for Skillfarm"""
        self.assertEqual(
            self.audit.get_esi_scopes(),
            ["esi-skills.read_skills.v1", "esi-skills.read_skillqueue.v1"],
        )

    def test_is_cooldown_should_return_false(self):
        """Test should return False for is_cooldown Property"""
        self.assertFalse(self.audit.is_cooldown)

    def test_is_cooldown_should_return_true(self):
        """Test should return True for is_cooldown Property"""
        self.audit.last_notification = timezone.now()
        self.assertTrue(self.audit.is_cooldown)

    def test_last_update_should_return_incomplete(self):
        """Test should return incomplete description for last_update Property"""
        self.assertEqual(
            self.audit.last_update, "One or more sections have not been updated"
        )

    def test_reset_has_token_error_should_return_false(self):
        """Test should reset has_token_error"""
        self.assertFalse(self.audit.reset_has_token_error())

    def test_reset_has_token_error_should_return_true(self):
        """Test should reset has_token_error"""
        create_update_status(
            self.audit,
            section=SkillFarmAudit.UpdateSection.SKILLQUEUE,
            is_success=False,
            error_message="",
            has_token_error=True,
            last_run_at=timezone.now(),
            last_run_finished_at=timezone.now(),
            last_update_at=timezone.now(),
            last_update_finished_at=timezone.now(),
        )
        self.assertTrue(self.audit.reset_has_token_error())



"""TestView class."""

# Standard Library
from unittest.mock import Mock, patch

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth
from allianceauth.authentication.models import UserProfile

# AA Skillfarm
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import create_user_from_evecharacter_with_access
from skillfarm.views import add_info_to_context

MODULE_PATH = "skillfarm.views."


class TestAddInfoContext(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )

    def test_add_info_context(self):
        request = self.factory.get(reverse("skillfarm:index"))
        request.user = self.user
        request.user.profile.theme = "dark"
        request.user.profile.save()
        context = {
            "page_title": "Skillfarm",
            "character_id": 1001,
        }
        response = add_info_to_context(request, context)

        new_context = {
            "theme": "dark",
            **context,
        }
        self.assertEqual(response, new_context)

    def test_add_info_context_no_userprofile(self):
        UserProfile.objects.filter(user=self.user).delete()
        request = self.factory.get(reverse("skillfarm:index"))
        request.user = self.user

        context = {
            "page_title": "Skillfarm",
            "character_id": 1001,
        }
        response = add_info_to_context(request, context)

        new_context = {
            "theme": None,
            **context,
        }
        self.assertEqual(response, new_context)



# Standard Library
from unittest.mock import patch

# Django
from django.test import TestCase

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from app_utils.testing import NoSocketsTestCase

# AA Skillfarm
from skillfarm import __title__
from skillfarm.decorators import (
    log_timing,
    when_esi_is_available,
)

DECORATOR_PATH = "skillfarm.decorators."


@patch(DECORATOR_PATH + "ESI_STATUS_ROUTE_RATE_LIMIT", new=1)
class TestDecorators(NoSocketsTestCase):
    @patch(DECORATOR_PATH + "fetch_esi_status")
    @patch(DECORATOR_PATH + "IS_TESTING", new=True)
    def test_when_esi_is_available_is_test(self, mock_fetch_esi_status):
        """Test when ESI is available in testing mode."""

        @when_esi_is_available
        def trigger_esi_deco():
            return "Testing Mode."

        # when
        result = trigger_esi_deco()
        # then
        self.assertEqual(result, "Testing Mode.")

    @patch(DECORATOR_PATH + "fetch_esi_status")
    @patch(DECORATOR_PATH + "IS_TESTING", new=False)
    def test_when_esi_is_ok(self, mock_fetch_esi_status):
        """Test when ESI is available in non-testing mode."""

        # ensure the fetch_esi_status call reports ESI as up
        mock_fetch_esi_status.return_value.is_ok = True

        @when_esi_is_available
        def trigger_esi_deco():
            return "Esi is Available"

        # when
        result = trigger_esi_deco()
        # then
        mock_fetch_esi_status.assert_called_once()
        self.assertEqual(result, "Esi is Available")

    @patch(DECORATOR_PATH + "fetch_esi_status")
    @patch(DECORATOR_PATH + "IS_TESTING", new=False)
    def test_when_esi_is_not_ok(self, mock_fetch_esi_status):
        """Test when ESI is not available in non-testing mode."""
        # ensure the fetch_esi_status call reports ESI as down
        mock_fetch_esi_status.return_value.is_ok = False

        @when_esi_is_available
        def trigger_esi_deco():
            return None

        # when
        result = trigger_esi_deco()
        # then
        self.assertIsNone(result)

    @patch(DECORATOR_PATH + "fetch_esi_status")
    @patch(DECORATOR_PATH + "IS_TESTING", new=False)
    def test_when_esi_is_ok_and_cached(self, mock_fetch_esi_status):
        """Test when ESI is available in non-testing mode and cached."""
        # Make a Exception to ensure fetch_esi_status is not called
        mock_fetch_esi_status.side_effect = Exception("Should not be called")

        @when_esi_is_available
        def trigger_esi_deco():
            return "Esi is Available"

        # when
        result = trigger_esi_deco()
        # then
        mock_fetch_esi_status.assert_not_called()
        self.assertEqual(result, "Esi is Available")

    def test_log_timing(self):
        # given
        logger = LoggerAddTag(get_extension_logger(__name__), __title__)

        @log_timing(logger)
        def trigger_log_timing():
            return "Log Timing"

        # when
        result = trigger_log_timing()
        # then
        self.assertEqual(result, "Log Timing")



# Generated by Django 4.2.20 on 2025-05-10 12:35

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("skillfarm", "0004_evetypeprice"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterskill",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="skillfarm_skills",
                to="skillfarm.skillfarmaudit",
            ),
        ),
        migrations.AlterField(
            model_name="characterskillqueueentry",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="skillfarm_skillqueue",
                to="skillfarm.skillfarmaudit",
            ),
        ),
        migrations.RemoveField(
            model_name="skillfarmaudit",
            name="last_update_skillqueue",
        ),
        migrations.RemoveField(
            model_name="skillfarmaudit",
            name="last_update_skills",
        ),
        migrations.CreateModel(
            name="CharacterUpdateStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        choices=[("skills", "Skills"), ("skillqueue", "Skill Queue")],
                        db_index=True,
                        max_length=32,
                    ),
                ),
                (
                    "is_success",
                    models.BooleanField(db_index=True, default=None, null=True),
                ),
                ("error_message", models.TextField()),
                ("has_token_error", models.BooleanField(default=False)),
                (
                    "last_run_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_run_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="skillfarm_update_status",
                        to="skillfarm.skillfarmaudit",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



# Standard Library
from io import StringIO
from unittest.mock import patch

# Django
from django.core.management import call_command
from django.db import IntegrityError, transaction
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm.models.prices import EveTypePrice
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse

COMMAND_PATH = "skillfarm.management.commands.skillfarm_load_prices"


@patch(COMMAND_PATH + ".requests.get")
@patch(COMMAND_PATH + ".EveType.objects.get_or_create_esi")
class TestLoadPrices(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

        cls.json = {
            1: {"buy": {"percentile": 100}, "sell": {"percentile": 200}},
            2: {"buy": {"percentile": 300}, "sell": {"percentile": 400}},
        }

    @patch(COMMAND_PATH + ".logger")
    def test_should_load_prices(self, mock_logger, _, mock_requests_get):
        # given
        mock_requests_get.return_value.json.return_value = self.json

        # when
        call_command("skillfarm_load_prices")
        # then
        mock_logger.debug.assert_called_once_with("Created all skillfarm prices.")

    @patch("builtins.input")
    def test_load_prices_should_get_integrityerror(
        self, mock_input, _, mock_requests_get
    ):
        # given
        EveTypePrice.objects.create(
            eve_type_id=1, buy=100, sell=200, updated_at=timezone.now()
        )

        mock_input.return_value = "n"
        mock_requests_get.return_value.json.return_value = self.json

        # when
        out = StringIO()
        with transaction.atomic():
            call_command("skillfarm_load_prices", stdout=out)
        output = out.getvalue()

        # then
        self.assertIn("No changes made.", output)
        excepted_count = EveTypePrice.objects.count()
        self.assertEqual(excepted_count, 1)

    @patch("builtins.input")
    def test_load_prices_should_get_integrityerror_and_replace(
        self, mock_input, _, mock_requests_get
    ):
        # given
        EveTypePrice.objects.create(
            eve_type_id=1, buy=100, sell=200, updated_at=timezone.now()
        )
        mock_input.return_value = "y"
        mock_requests_get.return_value.json.return_value = self.json

        # when
        out = StringIO()
        call_command("skillfarm_load_prices", stdout=out)
        output = out.getvalue()

        # then
        self.assertIn("Successfully updated 2 prices.", output)
        excepted_count = EveTypePrice.objects.count()
        self.assertEqual(excepted_count, 2)

    @patch(COMMAND_PATH + ".EveType.objects.get")
    def test_load_prices_should_evetype_not_exist(
        self, mock_evetype, _, mock_requests_get
    ):
        # given
        mock_requests_get.return_value.json.return_value = {
            666: {"buy": {"percentile": 100}, "sell": {"percentile": 200}},
        }
        mock_evetype.side_effect = EveType.DoesNotExist

        # when
        out = StringIO()
        call_command("skillfarm_load_prices", stdout=out)
        output = out.getvalue()

        self.assertIn("Ensure you have loaded the data from eveuniverse.", output)
        excepted_count = EveTypePrice.objects.count()
        self.assertEqual(excepted_count, 0)

    @patch(COMMAND_PATH + ".EveType.objects")
    def test_load_prices_should_get_error(self, mock_evetype, _, mock_requests_get):
        # given
        mock_requests_get.return_value.json.return_value = {
            666: {"buy": {"percentile": 100}, "sell": {"percentile": 200}},
        }
        mock_evetype.get.return_value = EveType.objects.get(id=44992)
        mock_evetype.filter.return_value.values_list.return_value = []

        # when
        out = StringIO()
        call_command("skillfarm_load_prices", stdout=out)
        output = out.getvalue()

        self.assertIn(
            "Error: Not all required types are loaded into the database.", output
        )
        excepted_count = EveTypePrice.objects.count()
        self.assertEqual(excepted_count, 0)



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm import __title__
from skillfarm.app_settings import SKILLFARM_BULK_METHODS_BATCH_SIZE
from skillfarm.decorators import log_timing
from skillfarm.providers import esi

if TYPE_CHECKING:
    # Alliance Auth
    from esi.stubs import CharactersCharacterIdSkillqueueGetItem as SkillQueueItems

    # AA Skillfarm
    from skillfarm.models.general import UpdateSectionResult
    from skillfarm.models.skillfarmaudit import (
        SkillFarmAudit,
    )

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class SkillqueueQuerySet(models.QuerySet):
    def finished_skills(self):
        """Return finished skills from a training queue."""
        return self.filter(
            finish_date__isnull=False,
            start_date__isnull=True,
            finish_date__gt=models.F("start_date"),
            finish_date__lt=timezone.now(),
            finished_level=5,
        )

    def extractions(self, character: "SkillFarmAudit") -> bool:
        """Return extraction ready skills from a training queue."""
        try:
            skillsetup = character.skillfarm_setup
            if not skillsetup or not skillsetup.skillset:
                skillset = []
            else:
                skillset = skillsetup.skillset
        except ObjectDoesNotExist:
            skillset = []

        extraction = self.filter(
            finish_date__gt=models.F("start_date"),
            finish_date__lt=timezone.now(),
            finished_level=5,
            eve_type__name__in=skillset,
        )

        return extraction

    def active_skills(self):
        """Return skills from an active training queue.
        Returns empty queryset when training is not active.
        """
        return self.filter(
            finish_date__isnull=False,
            start_date__isnull=False,
        )

    def skill_in_training(self):
        """Return current skill in training.
        Returns empty queryset when training is not active.
        """
        now_ = timezone.now()
        return self.active_skills().filter(
            start_date__lt=now_,
            finish_date__gt=now_,
        )

    def skill_filtered(self, character: "SkillFarmAudit") -> bool:
        """Return filtered skills from a training queue."""
        try:
            skillsetup = character.skillfarm_setup
            if not skillsetup or not skillsetup.skillset:
                skillset = []
            else:
                skillset = skillsetup.skillset
        except ObjectDoesNotExist:
            skillset = []

        skillqueue = self.filter(
            eve_type__name__in=skillset,
        )
        return skillqueue


class SkillqueueManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, character: "SkillFarmAudit", force_refresh: bool = False
    ) -> "UpdateSectionResult":
        """Update or Create skills for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.SKILLQUEUE,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(
        self, character: "SkillFarmAudit", force_refresh: bool = False
    ) -> dict:
        """Fetch Skillqueue entries from ESI data."""
        token = character.get_token()

        # Make the ESI request
        skillqueue_data = esi.client.Skills.GetCharactersCharacterIdSkillqueue(
            character_id=character.character.character_id,
            token=token,
        )
        character_skillqueue_items = skillqueue_data.results(
            force_refresh=force_refresh,
        )

        self._update_or_create_objs(
            character=character, character_skillqueue_items=character_skillqueue_items
        )

    @transaction.atomic()
    def _update_or_create_objs(
        self,
        character: "SkillFarmAudit",
        character_skillqueue_items: list["SkillQueueItems"],
    ) -> None:
        """Update or Create skill queue entries from objs data."""
        entries = []

        for entry in character_skillqueue_items:
            eve_type_instance, _ = EveType.objects.get_or_create_esi(id=entry.skill_id)
            entries.append(
                self.model(
                    name=character.name,
                    character=character,
                    eve_type=eve_type_instance,
                    finish_date=entry.finish_date,
                    finished_level=entry.finished_level,
                    level_end_sp=entry.level_end_sp,
                    level_start_sp=entry.level_start_sp,
                    queue_position=entry.queue_position,
                    start_date=entry.start_date,
                    training_start_sp=entry.training_start_sp,
                )
            )

        self.filter(character=character).delete()

        if len(entries) > 0:
            self.bulk_create(entries, batch_size=SKILLFARM_BULK_METHODS_BATCH_SIZE)


SkillqueueManager = SkillqueueManagerBase.from_queryset(SkillqueueQuerySet)



"""Shared ESI client for Skillfarm."""

# Alliance Auth
from esi.openapi_clients import ESIClientProvider

# AA Skillfarm
from skillfarm import (
    __app_name_useragent__,
    __characters_operations__,
    __esi_compatibility_date__,
    __github_url__,
    __title__,
    __version__,
)

esi = ESIClientProvider(
    compatibility_date=__esi_compatibility_date__,
    ua_appname=__app_name_useragent__,
    ua_version=__version__,
    ua_url=__github_url__,
    operations=__characters_operations__,
)



"""Initialize the app"""

__version__ = "1.0.2"
__title__ = "Skillfarm"

__package_name__ = "aa-skillfarm"
__app_name__ = "skillfarm"
__esi_compatibility_date__ = "2025-11-06"
__app_name_useragent__ = "AA-Skillfarm"

__github_url__ = f"https://github.com/Geuthur/{__package_name__}"

__characters_operations__ = [
    "GetCharactersCharacterIdSkills",
    "GetCharactersCharacterIdSkillqueue",
]



"""App URLs"""

# Django
from django.urls import path, re_path

# AA Skillfarm
from skillfarm import views
from skillfarm.api import api

app_name: str = "skillfarm"  # pylint: disable=invalid-name

urlpatterns = [
    # -- Views
    path("", views.index, name="index"),
    path("admin/", views.admin, name="admin"),
    path(
        "<int:character_id>/view/skillfarm/",
        views.skillfarm,
        name="skillfarm",
    ),
    path(
        "view/overview/",
        views.character_overview,
        name="character_overview",
    ),
    # -- Administration
    path("char/add/", views.add_char, name="add_char"),
    path(
        "switch_alarm/<int:character_id>/",
        views.switch_alarm,
        name="switch_alarm",
    ),
    path(
        "delete/<int:character_id>/",
        views.delete_character,
        name="delete_character",
    ),
    path(
        "skillset/<int:character_id>/",
        views.skillset,
        name="skillset",
    ),
    # -- Tools
    path("calc/", views.skillfarm_calc, name="calc"),
    # -- API System
    re_path(r"^api/", api.urls),
]



# Generated by Django 4.2.11 on 2024-10-22 21:21

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("skillfarm", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="skillfarmaudit",
            name="last_notification",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="skillfarmaudit",
            name="notification_sent",
            field=models.BooleanField(default=False),
        ),
        migrations.DeleteModel(
            name="SkillFarmNotification",
        ),
    ]



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Skillfarm
# AA Example App
from skillfarm import __version__


class SkillfarmConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    author = "Geuthur"
    name = "skillfarm"
    label = "skillfarm"
    verbose_name = f"Skillfarm v{__version__}"



# Django
from django.test import TestCase
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# AA Skillfarm
from skillfarm.models.skillfarmaudit import SkillFarmAudit
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.skillfarm import (
    add_skillfarmaudit_character_to_user,
    create_skillfarm_character,
    create_skillfarm_character_from_user,
    create_update_status,
    create_user_from_evecharacter,
    create_user_from_evecharacter_with_access,
)

MODULE_PATH = "skillfarm.managers.skillfarmaudit"


class TestCharacterAnnotateTotalUpdateStatus(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

    def test_should_be_ok(self):
        # given
        character = create_skillfarm_character(1001)
        sections = SkillFarmAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # when/then
        self.assertEqual(character.get_status, SkillFarmAudit.UpdateStatus.OK)

        # when
        qs = SkillFarmAudit.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, SkillFarmAudit.UpdateStatus.OK)

    def test_should_be_incomplete(self):
        # given
        character = create_skillfarm_character(1001)
        # when/then
        self.assertEqual(character.get_status, SkillFarmAudit.UpdateStatus.INCOMPLETE)

        # when
        qs = SkillFarmAudit.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(
            obj.total_update_status, SkillFarmAudit.UpdateStatus.INCOMPLETE
        )

    def test_should_be_token_error(self):
        # given
        character = create_skillfarm_character(1001)
        create_update_status(
            character,
            section=character.UpdateSection.SKILLS,
            is_success=False,
            error_message="",
            has_token_error=True,
            last_run_at=timezone.now(),
            last_run_finished_at=timezone.now(),
            last_update_at=timezone.now(),
            last_update_finished_at=timezone.now(),
        )
        # when/then
        self.assertEqual(character.get_status, SkillFarmAudit.UpdateStatus.TOKEN_ERROR)
        # when
        qs = SkillFarmAudit.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(
            obj.total_update_status, SkillFarmAudit.UpdateStatus.TOKEN_ERROR
        )

    def test_should_be_disabled(self):
        character = create_skillfarm_character(1001, active=False)
        # given
        sections = SkillFarmAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # then
        self.assertEqual(character.get_status, SkillFarmAudit.UpdateStatus.DISABLED)
        # when
        qs = SkillFarmAudit.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, SkillFarmAudit.UpdateStatus.DISABLED)

    def test_should_be_error(self):
        # given
        character = create_skillfarm_character(1001)
        sections = SkillFarmAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=False,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # then
        self.assertEqual(character.get_status, SkillFarmAudit.UpdateStatus.ERROR)
        # when
        qs = SkillFarmAudit.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, SkillFarmAudit.UpdateStatus.ERROR)


class TestSkillfarmAuditVisibleTo(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        cls.user, cls.characterownership = create_user_from_evecharacter_with_access(
            1001
        )

    def test_should_return_audit(self):
        # given
        character = create_skillfarm_character_from_user(self.user)
        # when
        qs = SkillFarmAudit.objects.visible_to(self.user)
        # then
        self.assertEqual(list(qs), [character])

    def test_should_return_empty_for_other_user(self):
        # given
        other_user, _ = create_user_from_evecharacter_with_access(1002)
        create_skillfarm_character_from_user(self.user)
        # when
        qs = SkillFarmAudit.objects.visible_to(other_user)
        # then
        self.assertEqual(list(qs), [])

    def test_should_return_multiple_audits_for_user_with_multiple_characters(self):
        # given
        character1 = create_skillfarm_character_from_user(self.user)
        character2 = add_skillfarmaudit_character_to_user(self.user, 1003)
        # when
        qs = SkillFarmAudit.objects.visible_to(self.user)
        # then
        self.assertCountEqual(list(qs), [character1, character2])

    def test_should_return_all_characters(self):
        # given
        other_user, _ = create_user_from_evecharacter(
            1002, permissions=["skillfarm.basic_access", "skillfarm.admin_access"]
        )
        character = create_skillfarm_character_from_user(self.user)
        character2 = create_skillfarm_character_from_user(other_user)
        # when
        qs = SkillFarmAudit.objects.visible_to(other_user)
        # then
        self.assertEqual(list(qs), [character, character2])


class TestSkillfarmAuditVisibleEveCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        cls.user, cls.characterownership = create_user_from_evecharacter_with_access(
            1001
        )

    def test_should_return_audit(self):
        # given
        create_skillfarm_character_from_user(self.user)
        eve_character = EveCharacter.objects.get(character_id=1001)
        # when
        qs = SkillFarmAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertEqual(list(qs), [eve_character])

    def test_should_return_multiple_audits_for_user_with_multiple_characters(self):
        # given
        create_skillfarm_character_from_user(self.user)
        add_skillfarmaudit_character_to_user(self.user, 1002)
        eve_character = EveCharacter.objects.get(character_id=1001)
        eve_character2 = EveCharacter.objects.get(character_id=1002)
        # when
        qs = SkillFarmAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertCountEqual(list(qs), [eve_character, eve_character2])

    def test_should_return_all_characters(self):
        # given
        other_user, _ = create_user_from_evecharacter(
            1002, permissions=["skillfarm.basic_access", "skillfarm.admin_access"]
        )

        eve_characters = EveCharacter.objects.all()
        # when
        qs = SkillFarmAudit.objects.visible_eve_characters(other_user)
        # then
        self.assertEqual(list(qs), list(eve_characters))



"""Models for Skillfarm."""

# Standard Library
import datetime
from collections.abc import Callable

# Third Party
from aiopenapi3.errors import HTTPClientError, HTTPServerError

# Django
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, Token
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.exceptions import HTTPNotModified

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm import __title__, app_settings
from skillfarm.managers.characterskill import SkillManager
from skillfarm.managers.skillfarmaudit import SkillFarmManager
from skillfarm.managers.skillqueue import SkillqueueManager
from skillfarm.models.general import UpdateSectionResult, _NeedsUpdate

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


# pylint: disable=too-many-public-methods
class SkillFarmAudit(models.Model):
    """Skillfarm Character Audit model"""

    class UpdateSection(models.TextChoices):
        SKILLS = "skills", _("Skills")
        SKILLQUEUE = "skillqueue", _("Skill Queue")

        @classmethod
        def get_sections(cls) -> list[str]:
            """Return list of section values."""
            return [choice.value for choice in cls]

        @property
        def method_name(self) -> str:
            """Return method name for this section."""
            return f"update_{self.value}"

    class UpdateStatus(models.TextChoices):
        DISABLED = "disabled", _("disabled")
        TOKEN_ERROR = "token_error", _("token error")
        ERROR = "error", _("error")
        OK = "ok", _("ok")
        INCOMPLETE = "incomplete", _("incomplete")
        IN_PROGRESS = "in_progress", _("in progress")

        def bootstrap_icon(self) -> str:
            """Return bootstrap corresponding icon class."""
            update_map = {
                status: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='skillfarm-tooltip' title='{self.description()}'>⬤</span>"
                )
                for status in [
                    self.DISABLED,
                    self.TOKEN_ERROR,
                    self.ERROR,
                    self.INCOMPLETE,
                    self.IN_PROGRESS,
                    self.OK,
                ]
            }
            return update_map.get(self, "")

        def bootstrap_text_style_class(self) -> str:
            """Return bootstrap corresponding bootstrap text style class."""
            update_map = {
                self.DISABLED: "text-muted",
                self.TOKEN_ERROR: "text-warning",
                self.INCOMPLETE: "text-warning",
                self.IN_PROGRESS: "text-info",
                self.ERROR: "text-danger",
                self.OK: "text-success",
            }
            return update_map.get(self, "")

        def description(self) -> str:
            """Return description for an enum object."""
            update_map = {
                self.DISABLED: _("Update is disabled"),
                self.TOKEN_ERROR: _("One section has a token error during update"),
                self.INCOMPLETE: _("One or more sections have not been updated"),
                self.IN_PROGRESS: _("Update is in progress"),
                self.ERROR: _("An error occurred during update"),
                self.OK: _("Updates completed successfully"),
            }
            return update_map.get(self, "")

    name = models.CharField(max_length=255, blank=True, null=True)

    active = models.BooleanField(default=True)

    character = models.OneToOneField(
        EveCharacter, on_delete=models.CASCADE, related_name="skillfarm_character"
    )

    notification = models.BooleanField(default=False)
    notification_sent = models.BooleanField(default=False)
    last_notification = models.DateTimeField(null=True, default=None, blank=True)

    objects = SkillFarmManager()

    def __str__(self):
        return f"{self.character.character_name} - Active: {self.active} - Status: {self.get_status}"

    class Meta:
        default_permissions = ()

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            "esi-skills.read_skills.v1",
            "esi-skills.read_skillqueue.v1",
        ]

    def get_token(self) -> Token:
        """Helper method to get a valid token for a specific character with specific scopes."""
        token = (
            Token.objects.filter(character_id=self.character.character_id)
            .require_scopes(self.get_esi_scopes())
            .require_valid()
            .first()
        )
        if token:
            return token
        return False

    @property
    def get_status(self) -> UpdateStatus.description:
        """Get the total update status of this character."""
        if self.active is False:
            return self.UpdateStatus.DISABLED

        qs = SkillFarmAudit.objects.filter(pk=self.pk).annotate_total_update_status()
        total_update_status = list(qs.values_list("total_update_status", flat=True))[0]
        return self.UpdateStatus(total_update_status)

    @property
    def notification_icon(self) -> str:
        """Get the notification icon for this character."""
        return format_html(
            render_to_string(
                "skillfarm/partials/icons/notification.html",
                {"status": self.notification},
            )
        )

    @property
    def last_update(self) -> UpdateStatus:
        """Get the last update status of this character."""
        return SkillFarmAudit.objects.last_update_status(self)

    @property
    def is_filtered(self) -> bool:
        """Check if the character has Skill Queue filter active."""
        return (
            self.skillfarm_skillqueue.skill_filtered(self).exists()
            or SkillFarmSetup.objects.filter(
                character=self,
                skillset__isnull=False,
            ).exists()
        )

    @property
    def is_skill_ready(self) -> bool:
        """Check if a character has skills for extraction."""
        return self.skillfarm_skills.extractions(self).exists()

    @property
    def is_skillqueue_ready(self) -> bool:
        """Check if a character has skillqueue ready for extraction."""
        return self.skillfarm_skillqueue.extractions(self).exists()

    @property
    def is_cooldown(self) -> bool:
        """Check if a character has a notification cooldown."""
        if (
            self.last_notification is not None
            and self.last_notification
            < timezone.now()
            - datetime.timedelta(days=app_settings.SKILLFARM_NOTIFICATION_COOLDOWN)
        ):
            return False
        if self.last_notification is None:
            return False
        return True

    @property
    def get_skillqueue(self) -> models.QuerySet["CharacterSkillqueueEntry"]:
        """Get the skillqueue for this character."""
        return self.skillfarm_skillqueue.all().select_related("eve_type")

    @property
    def get_skills(self) -> models.QuerySet["CharacterSkill"]:
        """Get the skills for this character."""
        return self.skillfarm_skills.all().select_related("eve_type")

    @property
    def get_skillsetup(self) -> models.QuerySet["SkillFarmSetup"] | None:
        """Get the skill setup for this character."""
        try:
            return self.skillfarm_setup
        except SkillFarmSetup.DoesNotExist:
            return None

    @property
    def extraction_icon(self) -> str:
        if self.is_skill_ready is True:
            return format_html(
                render_to_string("skillfarm/partials/icons/extraction_ready.html")
            )
        if self.is_skillqueue_ready is True:
            return format_html(
                render_to_string("skillfarm/partials/icons/extraction_sb_ready.html")
            )
        return ""

    def update_skills(self, force_refresh: bool = False) -> UpdateSectionResult:
        """Update skills for this character."""
        return self.skillfarm_skills.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    def update_skillqueue(self, force_refresh: bool = False) -> UpdateSectionResult:
        """Update skillqueue for this character."""
        return self.skillfarm_skillqueue.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    def calc_update_needed(self) -> _NeedsUpdate:
        """Calculate if an update is needed."""
        sections: models.QuerySet[CharacterUpdateStatus] = (
            self.skillfarm_update_status.all()
        )
        needs_update = {}
        for section in sections:
            needs_update[section.section] = section.need_update()
        return _NeedsUpdate(section_map=needs_update)

    def reset_update_status(self, section: UpdateSection) -> "CharacterUpdateStatus":
        """Reset the status of a given update section and return it."""
        update_status_obj: CharacterUpdateStatus = (
            self.skillfarm_update_status.get_or_create(
                section=section,
            )[0]
        )
        update_status_obj.reset()
        return update_status_obj

    def reset_has_token_error(self) -> bool:
        """Reset the has_token_error flag for this character."""
        update_status = self.get_status
        if update_status == self.UpdateStatus.TOKEN_ERROR:
            self.skillfarm_update_status.filter(
                has_token_error=True,
            ).update(
                has_token_error=False,
            )
            return True
        return False

    def update_section_if_changed(
        self,
        section: UpdateSection,
        fetch_func: Callable,
        force_refresh: bool = False,
    ):
        """Update the status of a specific section if it has changed."""
        section = self.UpdateSection(section)
        try:
            data = fetch_func(character=self, force_refresh=force_refresh)
            logger.debug("%s: Update has changed, section: %s", self, section.label)
        except HTTPNotModified as exc:
            logger.debug("%s: Update has not changed, section: %s", self, exc)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except HTTPServerError as exc:
            logger.debug("%s: Update has an HTTP internal server error: %s", self, exc)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except HTTPClientError as exc:
            error_message = f"{type(exc).__name__}: {str(exc)}"
            # TODO ADD DISCORD/AUTH NOTIFICATION?
            logger.error(
                "%s: %s: Update has Client Error: %s %s",
                self,
                section.label,
                error_message,
                exc.status_code,
            )
            return UpdateSectionResult(
                is_changed=False,
                is_updated=False,
                has_token_error=True,
                error_message=error_message,
            )
        return UpdateSectionResult(
            is_changed=True,
            is_updated=True,
            data=data,
        )

    def update_section_log(
        self,
        section: UpdateSection,
        result: UpdateSectionResult,
    ) -> None:
        """Update the status of a specific section."""
        error_message = result.error_message if result.error_message else ""
        is_success = not result.has_token_error
        defaults = {
            "is_success": is_success,
            "error_message": error_message,
            "has_token_error": result.has_token_error,
            "last_run_finished_at": timezone.now(),
        }
        obj: CharacterUpdateStatus = self.skillfarm_update_status.update_or_create(
            section=section,
            defaults=defaults,
        )[0]
        if result.is_updated:
            obj.last_update_at = obj.last_run_at
            obj.last_update_finished_at = timezone.now()
            obj.save()
        status = "successfully" if is_success else "with errors"
        logger.info("%s: %s Update run completed %s", self, section.label, status)

    def perform_update_status(
        self, section: UpdateSection, method: Callable, *args, **kwargs
    ) -> UpdateSectionResult:
        """Perform update status."""
        try:
            result = method(*args, **kwargs)
        except Exception as exc:
            error_message = f"{type(exc).__name__}: {str(exc)}"
            is_token_error = isinstance(exc, (TokenError))
            logger.error(
                "%s: %s: Error during update status: %s",
                self,
                section.label,
                error_message,
                exc_info=not is_token_error,  # do not log token errors
            )
            self.skillfarm_update_status.update_or_create(
                section=section,
                defaults={
                    "is_success": False,
                    "error_message": error_message,
                    "has_token_error": is_token_error,
                    "last_update_at": timezone.now(),
                },
            )
            raise exc
        return result

    def _generate_notification(self, skill_names: list[str]) -> str:
        """Generate notification for the user."""
        msg = _("%(charname)s: %(skillname)s") % {
            "charname": self.character.character_name,
            "skillname": ", ".join(skill_names),
        }
        return msg


class SkillFarmSetup(models.Model):
    """Skillfarm Character Skill Setup model for app"""

    id = models.AutoField(primary_key=True)

    name = models.CharField(max_length=255, blank=True, null=True)

    character = models.OneToOneField(
        SkillFarmAudit, on_delete=models.CASCADE, related_name="skillfarm_setup"
    )

    skillset = models.JSONField(default=dict, blank=True, null=True)

    def __str__(self):
        return f"{self.skillset}'s Skill Setup"

    objects = SkillFarmManager()

    class Meta:
        default_permissions = ()


class CharacterSkill(models.Model):
    """Skillfarm Character Skill model for app"""

    name = models.CharField(max_length=255, blank=True, null=True)

    character = models.ForeignKey(
        SkillFarmAudit, on_delete=models.CASCADE, related_name="skillfarm_skills"
    )
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    active_skill_level = models.PositiveIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)]
    )
    skillpoints_in_skill = models.PositiveBigIntegerField()
    trained_skill_level = models.PositiveBigIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)]
    )

    objects = SkillManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character}-{self.eve_type.name}"


class CharacterSkillqueueEntry(models.Model):
    """Skillfarm Skillqueue model for app"""

    name = models.CharField(max_length=255, blank=True, null=True)

    character = models.ForeignKey(
        SkillFarmAudit,
        on_delete=models.CASCADE,
        related_name="skillfarm_skillqueue",
    )

    queue_position = models.PositiveIntegerField(db_index=True)
    finish_date = models.DateTimeField(default=None, null=True)
    finished_level = models.PositiveIntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(5)]
    )
    level_end_sp = models.PositiveIntegerField(default=None, null=True)
    level_start_sp = models.PositiveIntegerField(default=None, null=True)
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    start_date = models.DateTimeField(default=None, null=True)
    training_start_sp = models.PositiveIntegerField(default=None, null=True)

    # TODO: Add to Notification System
    has_no_skillqueue = models.BooleanField(default=False)
    last_check = models.DateTimeField(default=None, null=True)

    objects = SkillqueueManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character}-{self.queue_position}"


class CharacterUpdateStatus(models.Model):
    """A Model to track the status of the last update."""

    character = models.ForeignKey(
        SkillFarmAudit, on_delete=models.CASCADE, related_name="skillfarm_update_status"
    )
    section = models.CharField(
        max_length=32, choices=SkillFarmAudit.UpdateSection.choices, db_index=True
    )
    is_success = models.BooleanField(default=None, null=True, db_index=True)
    error_message = models.TextField()
    has_token_error = models.BooleanField(default=False)

    last_run_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been started at this time",
    )
    last_run_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been successful finished at this time",
    )
    last_update_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been started at this time",
    )
    last_update_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been successful finished at this time",
    )

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character} - {self.section} - {self.is_success}"

    def need_update(self) -> bool:
        """Check if the update is needed."""
        if not self.is_success or not self.last_update_finished_at:
            needs_update = True
        else:
            section_time_stale = app_settings.SKILLFARM_STALE_TYPES.get(
                self.section, 60
            )
            stale = timezone.now() - timezone.timedelta(minutes=section_time_stale)
            needs_update = self.last_run_finished_at <= stale

        if needs_update and self.has_token_error:
            logger.info(
                "%s: Ignoring update because of token error, section: %s",
                self.character,
                self.section,
            )
            needs_update = False

        return needs_update

    def reset(self) -> None:
        """Reset this update status."""
        self.is_success = None
        self.error_message = ""
        self.has_token_error = False
        self.last_run_at = timezone.now()
        self.last_run_finished_at = None
        self.save()



# Django
from django.test import TestCase
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm.api.helpers.core import generate_progressbar_html
from skillfarm.api.helpers.skilldetails import (
    _calculate_sum_progress_bar,
    calculate_single_progress_bar,
)
from skillfarm.api.skillfarm import SkillFarmQueueSchema, get_skillqueue_data
from skillfarm.models.skillfarmaudit import CharacterSkillqueueEntry
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import create_skillfarm_character

MODULE_PATH = "skillfarm.api.helpers."


class Test_Calculate_Single_Progress_Bar(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.audit = create_skillfarm_character(1001)
        cls.skill1 = EveType.objects.get(name="skill1")
        cls.skill2 = EveType.objects.get(name="skill2")

    def test_calc_single_progress_bar_no_end_sp(self):
        """Test should return 0 if the skill has no end SP"""
        characterskillqueue = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now(),
            finish_date=timezone.now() + timezone.timedelta(days=3),
            level_end_sp=0,
        )

        self.assertEqual(calculate_single_progress_bar(characterskillqueue), 0)

    def test_calc_single_progress_bar_100_percent(self):
        """Test should return 100 if the skill is already finished"""
        characterskillqueue = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=2),
            finish_date=timezone.now() - timezone.timedelta(days=1),
            level_end_sp=1000,
        )

        self.assertEqual(calculate_single_progress_bar(characterskillqueue), 100)

    def test_calc_single_progress_bar_below_zero(self):
        """Test should return 100 if the skill is already finished"""
        characterskillqueue = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() + timezone.timedelta(days=1),
            finish_date=timezone.now() + timezone.timedelta(days=3),
            level_end_sp=1000,
        )

        self.assertEqual(calculate_single_progress_bar(characterskillqueue), 0)

    def test_calc_single_progress_bar(self):
        """Test should return 25 if the skill is 25% finished"""
        characterskillqueue = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=1),
            finish_date=timezone.now() + timezone.timedelta(days=3),
            level_end_sp=1000,
        )

        self.assertEqual(calculate_single_progress_bar(characterskillqueue), 25.0)


class Test_Calculate_Sum_Progress_bar(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.audit = create_skillfarm_character(1001)
        cls.skill1 = EveType.objects.get(name="skill1")
        cls.skill2 = EveType.objects.get(name="skill2")

    def test_calc_sum_progress_bar_no_skills(self):
        """Test should return 0 if there are no skills"""
        excepted_progressbar = generate_progressbar_html(0)
        self.assertEqual(_calculate_sum_progress_bar({}), excepted_progressbar)

    def test_calc_sum_progress_bar_no_end_sp(self):
        """Test should return 0.0% if the skill has no end SP"""

        characterskillqueue = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now(),
            finish_date=timezone.now() + timezone.timedelta(days=3),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue.save()

        skillqueue_response: list[SkillFarmQueueSchema] = []
        skillqueue = self.audit.skillfarm_skillqueue.filter(
            character=self.audit
        ).select_related("eve_type")

        for entry in skillqueue:
            skillqueue_response.append(get_skillqueue_data(entry))

        excepted_progressbar = generate_progressbar_html(0.0)

        self.assertEqual(
            _calculate_sum_progress_bar(skillqueue_response), excepted_progressbar
        )

    def test_calc_sum_progress_bar_partial_progress(self):
        """Test should return 50.0%"""
        characterskillqueue = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=1),
            finish_date=timezone.now() + timezone.timedelta(days=1),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue.save()

        skillqueue_response: list[SkillFarmQueueSchema] = []
        skillqueue = self.audit.skillfarm_skillqueue.filter(
            character=self.audit
        ).select_related("eve_type")

        for entry in skillqueue:
            skillqueue_response.append(get_skillqueue_data(entry))

        excepted_progressbar = generate_progressbar_html(50.0)

        self.assertEqual(
            _calculate_sum_progress_bar(skillqueue_response), excepted_progressbar
        )

    def test_calc_sum_progress_bar_multiple_skills(self):
        """Test should return 75.0%"""
        characterskillqueue1 = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=2),
            finish_date=timezone.now() - timezone.timedelta(days=1),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue1.save()

        characterskillqueue2 = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=2,
            eve_type=self.skill2,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=1),
            finish_date=timezone.now() + timezone.timedelta(days=1),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue2.save()

        skillqueue_response: list[SkillFarmQueueSchema] = []
        skillqueue = self.audit.skillfarm_skillqueue.filter(
            character=self.audit
        ).select_related("eve_type")

        for entry in skillqueue:
            skillqueue_response.append(get_skillqueue_data(entry))

        excepted_progressbar = generate_progressbar_html(75.0)

        self.assertEqual(
            _calculate_sum_progress_bar(skillqueue_response), excepted_progressbar
        )

    def test_calc_sum_progress_bar_multiple_skills_below_zero(self):
        """Test should return 0.0%"""
        characterskillqueue1 = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() + timezone.timedelta(days=1),
            finish_date=timezone.now() + timezone.timedelta(days=3),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue1.save()

        characterskillqueue2 = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=2,
            eve_type=self.skill2,
            finished_level=5,
            start_date=timezone.now() + timezone.timedelta(days=1),
            finish_date=timezone.now() + timezone.timedelta(days=3),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue2.save()

        skillqueue_response: list[SkillFarmQueueSchema] = []
        skillqueue = self.audit.skillfarm_skillqueue.filter(
            character=self.audit
        ).select_related("eve_type")

        for entry in skillqueue:
            skillqueue_response.append(get_skillqueue_data(entry))

        excepted_progressbar = generate_progressbar_html(0.0)

        self.assertEqual(
            _calculate_sum_progress_bar(skillqueue_response), excepted_progressbar
        )

    def test_calc_sum_progress_bar_with_nodate(self):
        """Test should return 50.0%"""
        characterskillqueue1 = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=1,
            eve_type=self.skill1,
            finished_level=5,
            start_date=timezone.now() - timezone.timedelta(days=2),
            finish_date=timezone.now() - timezone.timedelta(days=1),
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue1.save()

        characterskillqueue2 = CharacterSkillqueueEntry.objects.create(
            character=self.audit,
            queue_position=2,
            eve_type=self.skill2,
            finished_level=5,
            start_date=None,
            finish_date=None,
            level_start_sp=0,
            training_start_sp=100,
            level_end_sp=1000,
        )
        characterskillqueue2.save()

        skillqueue_response: list[SkillFarmQueueSchema] = []
        skillqueue = self.audit.skillfarm_skillqueue.filter(
            character=self.audit
        ).select_related("eve_type")

        for entry in skillqueue:
            skillqueue_response.append(get_skillqueue_data(entry))

        excepted_progressbar = generate_progressbar_html(50.0)

        self.assertEqual(
            _calculate_sum_progress_bar(skillqueue_response), excepted_progressbar
        )



# Django
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.safestring import mark_safe

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Skillfarm
from skillfarm import __title__
from skillfarm.models.skillfarmaudit import SkillFarmAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def arabic_number_to_roman(value) -> str:
    """Map to convert arabic to roman numbers (1 to 5 only)"""
    my_map = {0: "-", 1: "I", 2: "II", 3: "III", 4: "IV", 5: "V"}
    try:
        return my_map[value]
    except KeyError:
        return "-"


def get_main_character(request, character_id) -> tuple[bool, EveCharacter]:
    """Get Character and check permissions"""
    perms = True
    try:
        main_char = EveCharacter.objects.get(character_id=character_id)
    except EveCharacter.DoesNotExist:
        main_char = EveCharacter.objects.select_related(
            "character_ownership",
            "character_ownership__user__profile",
            "character_ownership__user__profile__main_character",
        ).get(character_id=request.user.profile.main_character.character_id)

    # check access
    visible = SkillFarmAudit.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        perms = False
    return perms, main_char


def get_character(request, character_id):
    """Get Character and check permissions"""
    perms = True
    try:
        character = SkillFarmAudit.objects.get(character__character_id=character_id)
    except SkillFarmAudit.DoesNotExist:
        return False, None

    # check access
    visible = SkillFarmAudit.objects.visible_to(request.user)
    if character not in visible:
        perms = False
    return perms, character


def get_alts_queryset(main_char, corporations=None):
    """Get all alts for a main character, optionally filtered by corporations."""
    try:
        linked_corporations = main_char.character_ownership.user.character_ownerships.all().select_related(
            "character_ownership"
        )

        if corporations:
            linked_corporations = linked_corporations.filter(
                character__corporation_id__in=corporations
            )

        linked_corporations = linked_corporations.values_list("character_id", flat=True)

        return EveCharacter.objects.filter(id__in=linked_corporations)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def generate_toggle_notification_button(character: SkillFarmAudit) -> mark_safe:
    """Generate a toggle notification button for the skillfarm"""
    return format_html(
        render_to_string(
            "skillfarm/partials/buttons/toggle-notification.html",
            {
                "character": character,
            },
        )
    )


def generate_edit_skillsetup_button(character: SkillFarmAudit) -> mark_safe:
    """Generate an edit skillsetup button for the skillfarm"""
    return format_html(
        render_to_string(
            "skillfarm/partials/buttons/edit-skillsetup.html",
            {
                "character": character,
            },
        )
    )


def generate_delete_character_button(character: SkillFarmAudit) -> mark_safe:
    """Generate a delete character button for the skillfarm"""
    return format_html(
        render_to_string(
            "skillfarm/partials/buttons/delete-character.html",
            {
                "character": character,
            },
        )
    )


def generate_skillinfo_button(character: SkillFarmAudit) -> mark_safe:
    """Generate a skillinfo button for the skillfarm"""
    return format_html(
        render_to_string(
            "skillfarm/partials/buttons/skillinfo.html",
            {
                "character": character,
            },
        )
    )


def generate_progressbar_html(progress) -> str:
    """Generate a progress bar HTML"""
    return format_html(
        render_to_string(
            "skillfarm/partials/progressbar.html",
            {
                "progress": f"{progress:.2f}",
            },
        )
    )


def generate_status_icon_html(character: SkillFarmAudit, size: int = 32) -> str:
    """Generate a status icon HTML"""
    return format_html(
        render_to_string(
            "skillfarm/partials/icons/status.html",
            {"character": character, "size": size},
        )
    )



"""TestView class."""

# Standard Library
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils import timezone

# AA Skillfarm
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import (
    create_evetypeprice,
    create_skillfarm_character,
    create_user_from_evecharacter_with_access,
)
from skillfarm.views import skillfarm_calc

MODULE_PATH = "skillfarm.views"


class TestSkillFarmCalculator(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

        cls.plex = create_evetypeprice(
            44992, buy=100, sell=200, updated_at=timezone.now()
        )
        cls.skillinjector = create_evetypeprice(
            40520, buy=300, sell=400, updated_at=timezone.now()
        )
        cls.extractor = create_evetypeprice(
            40519, buy=500, sell=600, updated_at=timezone.now()
        )

    def test_skillcalculator_should_view_calc(self):
        # given
        request = self.factory.get(reverse("skillfarm:calc"))
        request.user = self.user
        # when
        response = skillfarm_calc(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "PLEX")
        self.assertContains(response, "Skill Injector")
        self.assertContains(response, "Skill Extractor")

    def test_skillcalculator_should_view_calc_with_character_id(self):
        # given
        request = self.factory.get(reverse("skillfarm:calc"))
        request.user = self.user
        # when
        response = skillfarm_calc(request, character_id=1001)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "PLEX")
        self.assertContains(response, "Skill Injector")
        self.assertContains(response, "Skill Extractor")



"""Forms for app."""

# Django
from django import forms

# Alliance Auth (External Libs)
from eveuniverse.models import EveType


class ConfirmForm(forms.Form):
    """Form Confirms."""

    character_id = forms.CharField(
        widget=forms.HiddenInput(),
    )


class SkillSetForm(forms.Form):
    """Form SkillSet."""

    character_id = forms.CharField(
        widget=forms.HiddenInput(),
    )

    selected_skills = forms.CharField(
        required=False,
        widget=forms.HiddenInput(),
    )

    skills = forms.ModelMultipleChoiceField(
        queryset=EveType.objects.filter(eve_group__eve_category__id=16)
        .select_related("eve_group", "eve_group__eve_category")
        .order_by("name"),
        required=False,
        widget=forms.SelectMultiple(
            attrs={
                "class": "form-select",
                "id": "skillSetSelect",
            }
        ),
    )



# Standard Library
from typing import List, Optional, Tuple

# Django
from django.contrib.auth.models import User
from django.db.models import Q

# Alliance Auth
from allianceauth.authentication.backends import StateBackend
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

# Alliance Auth (External Libs)
from app_utils.testing import add_character_to_user
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm.models.prices import EveTypePrice
from skillfarm.models.skillfarmaudit import (
    CharacterSkill,
    CharacterUpdateStatus,
    SkillFarmAudit,
    SkillFarmSetup,
)


def create_character(eve_character: EveCharacter, **kwargs) -> SkillFarmAudit:
    """Create a Skillfarm Character from EveCharacter"""
    params = {"name": eve_character.character_name, "character": eve_character}
    params.update(kwargs)
    character = SkillFarmAudit(**params)
    character.save()
    return character


def create_update_status(
    character_audit: SkillFarmAudit, **kwargs
) -> CharacterUpdateStatus:
    """Create a Update Status for a Character Audit"""
    params = {
        "character": character_audit,
    }
    params.update(kwargs)
    update_status = CharacterUpdateStatus(**params)
    update_status.save()
    return update_status


def create_user_from_evecharacter_with_access(
    character_id: int, disconnect_signals: bool = True
) -> tuple[User, CharacterOwnership]:
    """Create user with access from an existing eve character and use it as main."""
    auth_character = EveCharacter.objects.get(character_id=character_id)
    username = StateBackend.iterate_username(auth_character.character_name)
    user = AuthUtils.create_user(username, disconnect_signals=disconnect_signals)
    user = AuthUtils.add_permission_to_user_by_name(
        "skillfarm.basic_access", user, disconnect_signals=disconnect_signals
    )
    character_ownership = add_character_to_user(
        user,
        auth_character,
        is_main=True,
        scopes=SkillFarmAudit.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )
    return user, character_ownership


def create_user_from_evecharacter(
    character_id: int,
    permissions: list[str] | None = None,
    scopes: list[str] | None = None,
) -> tuple[User, CharacterOwnership]:
    """Create new allianceauth user from EveCharacter object.

    Args:
        character_id: ID of eve character
        permissions: list of permission names, e.g. `"my_app.my_permission"`
        scopes: list of scope names
    """
    auth_character = EveCharacter.objects.get(character_id=character_id)
    user = AuthUtils.create_user(auth_character.character_name.replace(" ", "_"))
    character_ownership = add_character_to_user(
        user, auth_character, is_main=True, scopes=scopes
    )
    if permissions:
        for permission_name in permissions:
            user = AuthUtils.add_permission_to_user_by_name(permission_name, user)
    return user, character_ownership


def create_skillfarm_character_from_user(user: User, **kwargs) -> SkillFarmAudit:
    eve_character = user.profile.main_character
    if not eve_character:
        raise ValueError("User needs to have a main character.")

    kwargs.update({"eve_character": eve_character})
    return create_character(**kwargs)


def create_skillfarm_character(character_id: int, **kwargs) -> SkillFarmAudit:
    """Create a Audit Character from a existing EveCharacter"""

    _, character_ownership = create_user_from_evecharacter_with_access(
        character_id, disconnect_signals=True
    )
    return create_character(character_ownership.character, **kwargs)


def create_skillsetup_character(character_id: int, skillset: list) -> SkillFarmSetup:
    """Create a SkillSet for Skillfarm Audit Character"""
    audit = SkillFarmAudit.objects.get(
        character__character_id=character_id,
    )

    skillsetup = SkillFarmSetup(
        character=audit,
        skillset=skillset,
    )
    skillsetup.save()

    return skillsetup


def create_evetypeprice(evetype_id: int, **kwargs) -> EveType:
    params = {
        "eve_type": EveType.objects.get(id=evetype_id),
    }
    params.update(kwargs)
    price = EveTypePrice(**params)
    price.save()
    return price


def create_skill_character(
    character_id: int,
    evetype_id: int,
    skillpoints: int,
    active_level: int = 5,
    trained_level: int = 5,
) -> CharacterSkill:
    """Create a Skill for Skillfarm Audit Character"""
    audit = SkillFarmAudit.objects.get(
        character__character_id=character_id,
    )

    skill = CharacterSkill(
        character=audit,
        eve_type=EveType.objects.get(id=evetype_id),
        skillpoints_in_skill=skillpoints,
        active_skill_level=active_level,
        trained_skill_level=trained_level,
    )
    skill.save()

    return skill


def add_auth_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True
) -> CharacterOwnership:
    auth_character = EveCharacter.objects.get(character_id=character_id)
    return add_character_to_user(
        user,
        auth_character,
        is_main=False,
        scopes=SkillFarmAudit.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )


def add_skillfarmaudit_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True, **kwargs
) -> SkillFarmAudit:
    character_ownership = add_auth_character_to_user(
        user,
        character_id,
        disconnect_signals=disconnect_signals,
    )
    return create_character(character_ownership.character, **kwargs)



# Generated by Django 5.1.7 on 2025-03-18 13:21

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("skillfarm", "0003_characterskill_name_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveTypePrice",
            fields=[
                ("name", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "eve_type",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to="eveuniverse.evetype",
                    ),
                ),
                ("buy", models.DecimalField(decimal_places=2, max_digits=20)),
                ("sell", models.DecimalField(decimal_places=2, max_digits=20)),
                ("updated_at", models.DateTimeField()),
            ],
        ),
    ]



# Third Party
import requests

# Django
from django.core.management.base import BaseCommand
from django.db import IntegrityError, transaction
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm import __title__
from skillfarm.app_settings import SKILLFARM_PRICE_SOURCE_ID
from skillfarm.models.prices import EveTypePrice

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = "Preloads price data required for the skillfarm from Fuzzwork market API"

    # pylint: disable=unused-argument
    def handle(self, *args, **options):
        type_ids = []
        market_data = {}

        # Ensure that the required types are loaded into the database
        EveType.objects.get_or_create_esi(id=44992)
        EveType.objects.get_or_create_esi(id=40520)
        EveType.objects.get_or_create_esi(id=40519)

        # Get all skillfarm relevant ids
        typeids = EveType.objects.filter(id__in=[44992, 40520, 40519]).values_list(
            "id", flat=True
        )

        if len(typeids) != 3:
            self.stdout.write(
                "Error: Not all required types are loaded into the database."
            )
            return

        for item in typeids:
            type_ids.append(item)

        request = requests.get(
            "https://market.fuzzwork.co.uk/aggregates/",
            params={
                "types": ",".join([str(x) for x in type_ids]),
                "station": SKILLFARM_PRICE_SOURCE_ID,
            },
        ).json()

        market_data.update(request)

        # Create Bulk Object
        objs = []

        for key, value in market_data.items():
            try:
                eve_type = EveType.objects.get(id=key)

                item = EveTypePrice(
                    name=eve_type.name,
                    eve_type=eve_type,
                    buy=float(value["buy"]["percentile"]),
                    sell=float(value["sell"]["percentile"]),
                    updated_at=timezone.now(),
                )

                objs.append(item)
            except EveType.DoesNotExist:
                self.stdout.write(
                    f"EveType {key} not found. Skipping... Ensure you have loaded the data from eveuniverse."
                )
                continue

        try:
            with transaction.atomic():
                EveTypePrice.objects.bulk_create(objs)
                self.stdout.write(f"Successfully created {len(objs)} prices.")
                logger.debug("Created all skillfarm prices.")
                return
        except IntegrityError:
            self.stdout.write("Error: Prices already loaded into database.")
            update_arg = input("Would you like to update all prices? (y/n): ")

            if update_arg == "y":
                count = 0
                for obj in objs:
                    with transaction.atomic():
                        EveTypePrice.objects.update_or_create(
                            eve_type_id=obj.eve_type_id,
                            defaults={
                                "buy": obj.buy,
                                "sell": obj.sell,
                                "updated_at": obj.updated_at,
                            },
                        )
                    count += 1
                self.stdout.write(f"Successfully updated {count} prices.")
                logger.debug("Updated all skillfarm prices.")
            else:
                self.stdout.write("No changes made.")
            return



# Django
from django.contrib.admin.sites import AdminSite
from django.test import RequestFactory, TestCase

# AA Skillfarm
from skillfarm.admin import SkillFarmAuditAdmin
from skillfarm.models.skillfarmaudit import SkillFarmAudit
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.skillfarm import create_skillfarm_character

MODULE_PATH = "skillfarm.admin."


class TestAdminView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.adminmodel = SkillFarmAuditAdmin(
            model=SkillFarmAudit, admin_site=AdminSite()
        )
        cls.character = create_skillfarm_character(1001)
        cls.user = cls.character.character.character_ownership.user

    def test_column_entity_pic(self):
        """Test if the entity pic column is displayed correctly"""
        self.assertEqual(
            self.adminmodel._entity_pic(self.character),
            '<img src="https://images.evetech.net/characters/1001/portrait?size=32" class="img-circle">',
        )

    def test_column_character(self):
        """Test if the character column is displayed correctly"""
        self.assertEqual(self.adminmodel._character__character_id(self.character), 1001)

    def test_column_character_name(self):
        """Test if the character name column is displayed correctly"""
        self.assertEqual(
            self.adminmodel._character__character_name(self.character), "Gneuten"
        )

    def test_has_add_permission(self):
        """Test if the user has the permission to add should be false"""
        self.assertFalse(self.adminmodel.has_add_permission(RequestFactory()))

    def test_has_change_permission(self):
        """Test if the user has the permission to change should be false"""
        self.assertFalse(self.adminmodel.has_change_permission(RequestFactory()))



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Skillfarm
from skillfarm import __title__
from skillfarm.api.helpers.core import (
    generate_progressbar_html,
)
from skillfarm.models.skillfarmaudit import CharacterSkillqueueEntry

if TYPE_CHECKING:
    # AA Skillfarm
    from skillfarm.api.skillfarm import SkillFarmQueueSchema

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def calculate_single_progress_bar(skill: CharacterSkillqueueEntry):
    """Calculate the progress bar for a single skill"""
    totalsp = skill.level_end_sp
    start_date = skill.start_date
    finish_date = skill.finish_date

    if totalsp == 0:
        return 0

    current_date = timezone.now()
    total_duration = (finish_date - start_date).total_seconds()
    elapsed_duration = (current_date - start_date).total_seconds()

    if elapsed_duration > total_duration:
        progress = 100
    else:
        progress = (elapsed_duration / total_duration) * 100

    # Ensure the progress percentage is between 0 and 100
    progress = max(progress, 0)

    return round(progress, 2)


def _calculate_sum_progress_bar(
    skill_queue_response: list["SkillFarmQueueSchema"],
) -> str:
    """Calculate the progress bar for the skillqueue"""
    # Calculate the progress percentage for each skill individually
    total_progress_percent = 0
    skill_count = len(skill_queue_response)

    if skill_count == 0:
        return generate_progressbar_html(0)

    for skill in skill_queue_response:
        if skill.start_date and skill.finish_date == "-":
            continue
        total_progress_percent += skill.progress["value"]

    # Calculate the average progress percentage
    average_progress_percent = total_progress_percent / skill_count

    return generate_progressbar_html(average_progress_percent)



"""Model for Permission."""

# Standard Library
from dataclasses import dataclass
from typing import Any, NamedTuple

# Django
from django.db import models

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Skillfarm
from skillfarm import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class General(models.Model):
    """General model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app, Skillfarm."),
            ("corp_access", "Has access to all characters in the corporation."),
            ("admin_access", "Has access to all characters."),
        )


class UpdateSectionResult(NamedTuple):
    """A result of an attempted section update."""

    is_changed: bool | None
    is_updated: bool
    has_token_error: bool = False
    error_message: str | None = None
    data: Any = None


@dataclass(frozen=True)
class _NeedsUpdate:
    """An Object to track if an update is needed."""

    section_map: dict[str, bool]

    def __bool__(self) -> bool:
        """Check if any section needs an update."""
        return any(self.section_map.values())

    def for_section(self, section: str) -> bool:
        """Check if an update is needed for a specific section."""
        return self.section_map.get(section, False)



"""
App Settings
"""

# Standard Library
import sys

# Alliance Auth (External Libs)
from app_utils.app_settings import clean_setting

IS_TESTING = sys.argv[1:2] == ["test"]

# EVE Online Swagger
EVE_BASE_URL = "https://esi.evetech.net/"
EVE_API_URL = "https://esi.evetech.net/latest/"
EVE_BASE_URL_REGEX = r"^http[s]?:\/\/esi.evetech\.net\/"

# Fuzzwork
FUZZ_BASE_URL = "https://www.fuzzwork.co.uk/"
FUZZ_API_URL = "https://www.fuzzwork.co.uk/api/"
FUZZ_BASE_URL_REGEX = r"^http[s]?:\/\/(www\.)?fuzzwork\.co\.uk\/"

# ZKillboard
ZKILLBOARD_BASE_URL = "https://zkillboard.com/"
ZKILLBOARD_API_URL = "https://zkillboard.com/api/"
ZKILLBOARD_BASE_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/"
ZKILLBOARD_KILLMAIL_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/kill\/\d+\/"

# Set Test Mode True or False

# Set Naming on Auth Hook
SKILLFARM_APP_NAME = clean_setting("SKILLFARM_APP_NAME", "Skillfarm")

# If True you need to set up the Logger
SKILLFARM_LOGGER_USE = clean_setting("SKILLFARM_LOGGER_USE", False)

# Max Time to set Char Inactive
SKILLFARM_CHAR_MAX_INACTIVE_DAYS = clean_setting("SKILLFARM_CHAR_MAX_INACTIVE_DAYS", 3)

# Batch Size for Bulk Methods
SKILLFARM_BULK_METHODS_BATCH_SIZE = clean_setting(
    "SKILLFARM_BULK_METHODS_BATCH_SIZE", 500
)

# Set Notification Cooldown in Days
SKILLFARM_NOTIFICATION_COOLDOWN = clean_setting("SKILLFARM_NOTIFICATION_COOLDOWN", 3)

# Global timeout for tasks in seconds to reduce task accumulation during outages.
SKILLFARM_TASKS_TIME_LIMIT = clean_setting("SKILLFARM_TASKS_TIME_LIMIT", 7200)

# Price Source default is Jita
SKILLFARM_PRICE_SOURCE_ID = clean_setting("SKILLFARM_PRICE_SOURCE_ID", 60003760)

SKILLFARM_STALE_TYPES = {
    "skills": 30,
    "skillqueue": 30,
}



"""TestView class."""

# Standard Library
import json
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse

# AA Skillfarm
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import (
    create_skillfarm_character,
    create_user_from_evecharacter_with_access,
)
from skillfarm.views import switch_alarm

MODULE_PATH = "skillfarm.views"


class TestSwitchalarmView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.audit = create_skillfarm_character(1001)
        cls.user = cls.audit.character.character_ownership.user
        cls.no_audit_user, _ = create_user_from_evecharacter_with_access(1002)

    def test_switch_alarm(self):
        character_id = self.audit.character.character_id
        form_data = {
            "character_id": character_id,
            "confirm": "yes",
        }

        request = self.factory.post(
            reverse("skillfarm:switch_alarm", args=[character_id]), data=form_data
        )
        request.user = self.user

        response = switch_alarm(request, character_id=character_id)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTrue(response_data["success"])
        self.assertEqual(response_data["message"], "Alarm successfully updated")

    def test_switch_alarm_no_audit(self):
        form_data = {
            "character_id": 1001,
            "confirm": "yes",
        }

        request = self.factory.post(
            reverse("skillfarm:switch_alarm", args=[1001]), data=form_data
        )
        request.user = self.no_audit_user

        response = switch_alarm(request, character_id=1001)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)
        self.assertFalse(response_data["success"])
        self.assertEqual(response_data["message"], "Permission Denied")

    def test_switch_alarm_invalid(self):
        request = self.factory.post(
            reverse("skillfarm:switch_alarm", args=[1001]), data=None
        )
        request.user = self.no_audit_user

        response = switch_alarm(request, character_id=1001)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.BAD_REQUEST)
        self.assertFalse(response_data["success"])
        self.assertEqual(response_data["message"], "Invalid Form")



"""
Constants
"""

# Standard Library
import os

AA_SKILLFARM_BASE_DIR = os.path.join(os.path.dirname(__file__))
AA_SKILLFARM_STATIC_DIR = os.path.join(AA_SKILLFARM_BASE_DIR, "static", "skillfarm")

# Embed colors
DISCORD_EMBED_COLOR_INFO = 0x5BC0DE
DISCORD_EMBED_COLOR_SUCCESS = 0x5CB85C
DISCORD_EMBED_COLOR_WARNING = 0xF0AD4E
DISCORD_EMBED_COLOR_DANGER = 0xD9534F

# Discord embed color map
DISCORD_EMBED_COLOR_MAP = {
    "info": DISCORD_EMBED_COLOR_INFO,
    "success": DISCORD_EMBED_COLOR_SUCCESS,
    "warning": DISCORD_EMBED_COLOR_WARNING,
    "danger": DISCORD_EMBED_COLOR_DANGER,
}



"""TestView class."""

# Standard Library
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testdata_factories import UserMainFactory

# AA Skillfarm
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import create_user_from_evecharacter_with_access
from skillfarm.views import admin, character_overview, index, skillfarm, skillfarm_calc

MODULE_PATH = "skillfarm.views."


class TestViewAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.no_evecharacter_user = UserMainFactory(
            permissions=[
                "skillfarm.basic_access",
            ]
        )

    def test_index(self):
        # given
        request = self.factory.get(reverse("skillfarm:index"))
        request.user = self.user
        # when
        response = index(request)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(
            response.url,
            reverse(
                "skillfarm:skillfarm",
                args=[self.user.profile.main_character.character_id],
            ),
        )

    def test_skillfarm(self):
        # given
        request = self.factory.get(
            reverse(
                "skillfarm:skillfarm",
                args=[self.user.profile.main_character.character_id],
            )
        )
        request.user = self.user
        # when
        response = skillfarm(
            request, character_id=self.user.profile.main_character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Skillfarm")

    def test_skillfarm_no_character(self):
        # given
        request = self.factory.get(
            reverse(
                "skillfarm:skillfarm",
                args=[self.user.profile.main_character.character_id],
            )
        )
        request.user = self.no_evecharacter_user
        # when
        response = skillfarm(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Skillfarm")

    def test_character_overview(self):
        # given
        request = self.factory.get(
            reverse(
                "skillfarm:character_overview",
            )
        )
        request.user = self.user
        # when
        response = character_overview(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Overview")

    def test_skill_calculator(self):
        # given
        request = self.factory.get(reverse("skillfarm:calc"))
        request.user = self.user
        # when
        response = skillfarm_calc(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(
            response, "An error occurred while fetching the market data."
        )

    def test_admin(self):
        # given
        self.user.is_superuser = True
        self.user.save()

        request = self.factory.get(reverse("skillfarm:admin"))
        request.user = self.user
        # when
        response = admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Administration")



# Django
from django.test import RequestFactory, TestCase

# Alliance Auth (External Libs)
from app_utils.testdata_factories import UserMainFactory
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm.api.helpers import core
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import (
    create_skillfarm_character,
    create_user_from_evecharacter_with_access,
)

MODULE_PATH = "skillfarm.api.helpers."


class TestCoreHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.audit = create_skillfarm_character(1001)
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.user_2, cls.character_ownership_2 = (
            create_user_from_evecharacter_with_access(1002)
        )
        cls.no_evecharacter_user = UserMainFactory(
            permissions=[
                "skillfarm.basic_access",
            ]
        )

    def test_generate_progressbar_html(self):
        """Test the generate_progressbar_html function."""
        result = core.generate_progressbar_html(50)
        self.assertIn("width: 50.00%;", result)
        self.assertIn("50.00%", result)

    def test_get_main_character(self):
        """Test getting the main character for a user."""
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        perm, main_character = core.get_main_character(
            request, self.character_ownership.character.character_id
        )
        # then
        self.assertEqual(
            main_character.character_id, self.character_ownership.character.character_id
        )
        self.assertTrue(perm)

    def test_get_main_character_no_permission(self):
        """Test getting a character without permission."""
        # given
        request = self.factory.get("/")
        request.user = self.no_evecharacter_user
        # when
        perm, main_character = core.get_main_character(
            request, self.character_ownership.character.character_id
        )
        # then
        self.assertFalse(perm)
        self.assertEqual(
            main_character.character_id, self.character_ownership.character.character_id
        )

    def test_get_main_character_nonexistent(self):
        """Test getting a nonexistent character."""
        # given
        request = self.factory.get("/")
        request.user = self.user_2
        # when
        perm, main_character = core.get_main_character(request, 999999999)
        # then
        self.assertTrue(perm)
        self.assertEqual(
            main_character.character_id, 1002
        )  # Is the main character of user_2



# Third Party
from charlink.app_imports.utils import AppImport, LoginImport

# Django
from django.contrib.auth.models import Permission
from django.db.models import Exists, OuterRef

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.allianceauth import users_with_permission

# AA Skillfarm
from skillfarm.app_settings import SKILLFARM_APP_NAME
from skillfarm.models.skillfarmaudit import SkillFarmAudit
from skillfarm.tasks import update_character


# pylint: disable=unused-argument, duplicate-code
def _add_character_charaudit(request, token):
    skillfarm = SkillFarmAudit.objects.update_or_create(
        character=EveCharacter.objects.get_character_by_id(token.character_id),
        defaults={"name": token.character_name},
    )[0]

    update_character.apply_async(
        args=[skillfarm.pk], kwargs={"force_refresh": True}, priority=6
    )


def _is_character_added_charaudit(character: EveCharacter):
    return SkillFarmAudit.objects.filter(character=character).exists()


def _users_with_perms_charaudit():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label="skillfarm", codename="basic_access"
        )
    )


app_import = AppImport(
    "skillfarm",
    [
        LoginImport(
            app_label="skillfarm",
            unique_id="default",
            field_label=SKILLFARM_APP_NAME,
            add_character=_add_character_charaudit,
            scopes=SkillFarmAudit.get_esi_scopes(),
            check_permissions=lambda user: user.has_perm("skillfarm.basic_access"),
            is_character_added=_is_character_added_charaudit,
            is_character_added_annotation=Exists(
                SkillFarmAudit.objects.filter(character_id=OuterRef("pk"))
            ),
            get_users_with_perms=_users_with_perms_charaudit,
        ),
    ],
)



"""Model for Prices."""

# Django
from django.db import models

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveType

# AA Skillfarm
from skillfarm import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveTypePrice(models.Model):
    name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
    )
    eve_type = models.OneToOneField(
        EveType,
        on_delete=models.deletion.CASCADE,
        primary_key=True,
    )
    buy = models.DecimalField(max_digits=20, decimal_places=2)
    sell = models.DecimalField(max_digits=20, decimal_places=2)
    updated_at = models.DateTimeField()



# Generated by Django 4.2.20 on 2025-05-04 21:51

# Django
import django.db.models.deletion
from django.db import migrations, models


def create_update_status(apps, schema_editor):
    CharacterUpdateStatus = apps.get_model("skillfarm", "CharacterUpdateStatus")
    SkillFarmAudit = apps.get_model("skillfarm", "SkillFarmAudit")

    for character in SkillFarmAudit.objects.all():
        sections = ["skills", "skillqueue"]
        for section in sections:
            update_status = CharacterUpdateStatus.objects.get_or_create(
                character=character,
                section=section,
            )[0]
            update_status.is_success = None
            update_status.error_message = ""
            update_status.has_token_error = False
            update_status.last_run_at = None
            update_status.last_run_finished_at = None
            update_status.save()


class Migration(migrations.Migration):

    dependencies = [
        ("skillfarm", "0005_alter_characterskill_character_and_more"),
    ]

    operations = [
        migrations.RunPython(create_update_status),
    ]



# Django
from django.contrib import admin
from django.contrib.humanize.templatetags.humanize import naturaltime
from django.db.models import Max, Q
from django.utils import timezone
from django.utils.html import format_html
from django.utils.timesince import timesince
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver

# AA Skillfarm
from skillfarm.models.skillfarmaudit import CharacterUpdateStatus, SkillFarmAudit


class CharacterUpdateStatusAdminInline(admin.TabularInline):
    model = CharacterUpdateStatus
    fields = (
        "section",
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "error_message",
        "run_finished_at",
        "_run_duration",
        "update_finished_at",
        "_update_duration",
    )
    readonly_fields = (
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "_run_duration",
        "_update_duration",
    )
    ordering = ["section"]

    # pylint: disable=unused-argument
    def has_add_permission(self, request, obj=None):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False

    # pylint: disable=unused-argument
    def has_delete_permission(self, request, obj=None):
        return False

    @admin.display(boolean=True)
    def _is_enabled(self, obj: CharacterUpdateStatus) -> bool:
        return obj.is_enabled

    @admin.display(boolean=True)
    def _is_success(self, obj: CharacterUpdateStatus) -> bool:
        if not obj.is_enabled:
            return None
        return obj.is_success

    @admin.display(boolean=True)
    def _is_token_ok(self, obj: CharacterUpdateStatus) -> bool:
        return not obj.has_token_error

    @admin.display
    def _run_duration(self, obj: CharacterUpdateStatus) -> float:
        return self._calc_duration(obj.last_run_at, obj.last_run_finished_at)

    @admin.display
    def _update_duration(self, obj: CharacterUpdateStatus) -> float:
        return self._calc_duration(obj.last_update_at, obj.last_update_finished_at)

    @staticmethod
    def _calc_duration(
        started_at: timezone.datetime, finished_at: timezone.datetime
    ) -> str:
        if not started_at or not finished_at:
            return "-"
        return timesince(finished_at - started_at)


# TODO make a ETAG Cache clear option?
@admin.register(SkillFarmAudit)
class SkillFarmAuditAdmin(admin.ModelAdmin):
    """Admin Interface for Characters"""

    model = SkillFarmAudit
    model._meta.verbose_name = "Character"
    model._meta.verbose_name_plural = "Characters"

    list_display = (
        "_entity_pic",
        "_character__character_id",
        "_character__character_name",
        "_last_update_at",
    )

    list_display_links = (
        "_entity_pic",
        "_character__character_id",
        "_character__character_name",
    )

    list_select_related = ("character",)

    ordering = ["character__character_name"]

    search_fields = ["character__character_name", "character__character_id"]

    actions = [
        "delete_objects",
    ]

    inlines = (CharacterUpdateStatusAdminInline,)

    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.prefetch_related("skillfarm_update_status").annotate(
            last_update_at=Max(
                "skillfarm_update_status__last_run_finished_at",
                filter=~Q(skillfarm_update_status__section="skillqueue"),
            )
        )

    @admin.display(description="")
    def _entity_pic(self, obj: SkillFarmAudit):
        eve_id = obj.character.character_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("character", eve_id, 32),
        )

    @admin.display(description="Character ID", ordering="character__character_id")
    def _character__character_id(self, obj: SkillFarmAudit):
        return obj.character.character_id

    @admin.display(description="Character Name", ordering="character__character_name")
    def _character__character_name(self, obj: SkillFarmAudit):
        return obj.character.character_name

    @admin.display(ordering="last_update_at", description=_("last update run"))
    def _last_update_at(self, obj: SkillFarmAudit):
        return naturaltime(obj.last_update_at) if obj.last_update_at else "-"

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False



"""PvE Views"""

# Standard Library
import json

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

# Alliance Auth
from allianceauth.authentication.models import UserProfile
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Skillfarm
from skillfarm import __title__, forms, tasks
from skillfarm.api.helpers.core import get_character
from skillfarm.models.prices import EveTypePrice
from skillfarm.models.skillfarmaudit import SkillFarmAudit, SkillFarmSetup
from skillfarm.tasks import update_all_skillfarm

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


# pylint: disable=unused-argument
def add_info_to_context(request, context: dict) -> dict:
    """Add additional information to the context for the view."""
    theme = None
    try:
        user = UserProfile.objects.get(id=request.user.id)
        theme = user.theme
    except UserProfile.DoesNotExist:
        pass

    new_context = {
        **{"theme": theme},
        **context,
    }
    return new_context


@login_required
@permission_required("skillfarm.basic_access")
def index(request):
    """Index View"""
    return redirect(
        "skillfarm:skillfarm", request.user.profile.main_character.character_id
    )


@login_required
@permission_required("skillfarm.basic_access")
def admin(request):
    """Admin View"""
    context = {
        "page_title": "Admin",
    }

    if not request.user.is_superuser:
        messages.error(request, _("You do not have permission to access this page."))
        return redirect("skillfarm:index")

    if request.method == "POST":
        force_refresh = False
        if request.POST.get("force_refresh", False):
            force_refresh = True
        if request.POST.get("run_char_updates"):
            messages.info(request, _("Queued Update All Characters"))
            update_all_skillfarm.apply_async(
                kwargs={"force_refresh": force_refresh}, priority=7
            )
    return render(request, "skillfarm/admin.html", context=context)


@login_required
@permission_required("skillfarm.basic_access")
def skillfarm(request, character_id=None):
    """Main Skillfarm View"""
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    context = {
        "page_title": "Skillfarm",
        "character_id": character_id,
        "forms": {
            "confirm": forms.ConfirmForm(),
            "skillset": forms.SkillSetForm(),
        },
    }
    context = add_info_to_context(request, context)
    return render(request, "skillfarm/skillfarm.html", context=context)


@login_required
@permission_required("skillfarm.basic_access")
def character_overview(request):
    """Character Overview"""
    context = {
        "page_title": "Character Overview",
    }
    context = add_info_to_context(request, context)

    return render(request, "skillfarm/overview.html", context=context)


@login_required
@token_required(scopes=SkillFarmAudit.get_esi_scopes())
@permission_required("skillfarm.basic_access")
def add_char(request, token):
    """Add Character to Skillfarm"""
    character = EveCharacter.objects.get_character_by_id(token.character_id)
    char = SkillFarmAudit.objects.update_or_create(
        character=character, defaults={"name": token.character_name}
    )[0]
    tasks.update_character.apply_async(args=[char.pk], kwargs={"force_refresh": True})

    msg = _(
        "{character_name} successfully added or updated to Skillfarm System"
    ).format(
        character_name=char.character.character_name,
    )
    messages.success(request, msg)
    return redirect("skillfarm:index")


@login_required
@permission_required("skillfarm.basic_access")
@require_POST
def switch_alarm(request, character_id: int):
    """Switch Character Notification Alarm"""
    # Check Permission & If Character Exists
    perm, __ = get_character(request, character_id)
    form = forms.ConfirmForm(request.POST)
    if form.is_valid():
        if not perm:
            msg = _("Permission Denied")
            return JsonResponse(
                {"success": False, "message": msg}, status=403, safe=False
            )

        character_id = form.cleaned_data["character_id"]

        character = SkillFarmAudit.objects.get(character__character_id=character_id)
        character.notification = not character.notification
        character.save()
        msg = _("Alarm successfully updated")
    else:
        msg = "Invalid Form"
        return JsonResponse({"success": False, "message": msg}, status=400, safe=False)
    return JsonResponse({"success": True, "message": msg}, status=200, safe=False)


@login_required
@permission_required("skillfarm.basic_access")
@require_POST
def delete_character(request, character_id: int):
    """Delete Character"""
    # Check Permission & If Character Exists
    perm, __ = get_character(request, character_id)
    form = forms.ConfirmForm(request.POST)
    if form.is_valid():
        if not perm:
            msg = _("Permission Denied")
            return JsonResponse(
                {"success": False, "message": msg}, status=403, safe=False
            )

        character_id = form.cleaned_data["character_id"]

        character = SkillFarmAudit.objects.get(character__character_id=character_id)
        character.delete()
        msg = _("{character_name} successfully deleted").format(
            character_name=character.character.character_name,
        )
    else:
        msg = "Invalid Form"
        return JsonResponse({"success": False, "message": msg}, status=400, safe=False)
    return JsonResponse({"success": True, "message": msg}, status=200, safe=False)


@login_required
@permission_required("skillfarm.basic_access")
@require_POST
def skillset(request, character_id: list):
    """Edit Character SkillSet"""
    # Check Permission & If Character Exists
    perm, __ = get_character(request, character_id)
    form = forms.SkillSetForm(request.POST)

    if form.is_valid():
        if not perm:
            msg = _("Permission Denied")
            return JsonResponse(
                {"success": False, "message": msg}, status=403, safe=False
            )
        character_id = form.cleaned_data["character_id"]
        selected_skills = form.cleaned_data["selected_skills"]

        character = SkillFarmAudit.objects.get(character__character_id=character_id)

        try:
            skills_data = json.loads(selected_skills)
            filtered = [
                entry["value"]
                for entry in skills_data
                if entry.get("selected") is True and entry.get("value")
            ]
            skillset_list = filtered if filtered else None
            SkillFarmSetup.objects.update_or_create(
                character=character, defaults={"skillset": skillset_list}
            )
        except Exception:  # pylint: disable=broad-except
            msg = _("Invalid JSON format")
            return JsonResponse(
                {"success": False, "message": msg}, status=400, safe=False
            )

        msg = _("{character_name} Skillset successfully updated").format(
            character_name=character.character.character_name,
        )
    else:
        msg = "Invalid Form"
        return JsonResponse({"success": False, "message": msg}, status=400, safe=False)
    return JsonResponse({"success": True, "message": msg}, status=200, safe=False)


@login_required
@permission_required("skillfarm.basic_access")
def skillfarm_calc(request, character_id=None):
    """Skillfarm Calc View"""
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    skillfarm_dict = {}
    error = False
    try:
        plex = EveTypePrice.objects.get(eve_type__id=44992)
        injector = EveTypePrice.objects.get(eve_type__id=40520)
        extractor = EveTypePrice.objects.get(eve_type__id=40519)

        plex_price = float(plex.sell)
        injector_price = float(injector.sell)
        extractor_price = float(extractor.sell)

        monthcalc = (injector_price * 3.5) - (
            (plex_price * 500) + (extractor_price * 3.5)
        )
        month12calc = (injector_price * 3.5) - (
            (plex_price * 300) + (extractor_price * 3.5)
        )
        month24calc = (injector_price * 3.5) - (
            (plex_price * 275) + (extractor_price * 3.5)
        )

        skillfarm_dict["plex"] = plex
        skillfarm_dict["injektor"] = injector
        skillfarm_dict["extratkor"] = extractor

        skillfarm_dict["calc"] = {
            "month": monthcalc,
            "month12": month12calc,
            "month24": month24calc,
        }
    except EveTypePrice.DoesNotExist:
        error = True

    context = {
        "error": {
            "status": error,
            "message": _(
                "An error occurred while fetching the market data. Please inform an admin to fetch Market Data."
            ),
        },
        "character_id": character_id,
        "page_title": "Skillfarm Calc",
        "skillfarm": skillfarm_dict,
    }

    return render(request, "skillfarm/calculator.html", context=context)



# Generated by Django 4.2.11 on 2024-10-21 21:42

# Django
import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app, Skillfarm."),
                    ("corp_access", "Has access to all characters in the corporation."),
                    ("admin_access", "Has access to all characters."),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="SkillFarmAudit",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("active", models.BooleanField(default=True)),
                ("notification", models.BooleanField(default=False)),
                (
                    "last_update_skills",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "last_update_skillqueue",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="skillfarm_character",
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="SkillFarmSetup",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("skillset", models.JSONField(blank=True, default=dict, null=True)),
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="skillfarm_setup",
                        to="skillfarm.skillfarmaudit",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="SkillFarmNotification",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("message", models.TextField()),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="skillfarm_notification",
                        to="skillfarm.skillfarmaudit",
                    ),
                ),
            ],
            options={
                "verbose_name": "Skillfarm Notification",
                "verbose_name_plural": "Skillfarm Notifications",
                "ordering": ["-timestamp"],
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterSkillqueueEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("queue_position", models.PositiveIntegerField(db_index=True)),
                ("finish_date", models.DateTimeField(default=None, null=True)),
                (
                    "finished_level",
                    models.PositiveIntegerField(
                        validators=[
                            django.core.validators.MinValueValidator(1),
                            django.core.validators.MaxValueValidator(5),
                        ]
                    ),
                ),
                ("level_end_sp", models.PositiveIntegerField(default=None, null=True)),
                (
                    "level_start_sp",
                    models.PositiveIntegerField(default=None, null=True),
                ),
                ("start_date", models.DateTimeField(default=None, null=True)),
                (
                    "training_start_sp",
                    models.PositiveIntegerField(default=None, null=True),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="skillqueue",
                        to="skillfarm.skillfarmaudit",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterSkill",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "active_skill_level",
                    models.PositiveIntegerField(
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(5),
                        ]
                    ),
                ),
                ("skillpoints_in_skill", models.PositiveBigIntegerField()),
                (
                    "trained_skill_level",
                    models.PositiveBigIntegerField(
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(5),
                        ]
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="character_skills",
                        to="skillfarm.skillfarmaudit",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



"""This module provides lazy loading of some common functions and objects that are not needed for every request."""

# Django
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import (
    character_portrait_url,
    type_render_url,
)


def get_character_portrait_url(
    character_id: int, size: int = 32, character_name: str = None, as_html: bool = False
) -> str:
    """Get the character portrait for a character ID."""

    render_url = character_portrait_url(character_id=character_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="character-portrait rounded-circle" src="{}" alt="{}">',
            render_url,
            character_name,
        )
        return render_html
    return render_url


def get_type_render_url(
    type_id: int, size: int = 32, type_name: str = None, as_html: bool = False
) -> str:
    """Get the type render for a type ID."""

    render_url = type_render_url(type_id=type_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="type-render rounded-circle" src="{}" alt="{}">',
            render_url,
            type_name,
        )
        return render_html
    return render_url


def str_normalize_time(evetime, hours: bool = False) -> str:
    """Normalize eve time to a string."""
    if hours:
        return timezone.localtime(evetime).strftime("%Y-%m-%d %H:%M")
    return timezone.localtime(evetime).strftime("%Y-%m-%d")



"""Discord helper functions"""

# Third Party
from celery import shared_task

# Django
from django.apps import apps
from django.utils import timezone

# Alliance Auth
from allianceauth.authentication.models import User
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Skillfarm
from skillfarm import __title__
from skillfarm.constants import DISCORD_EMBED_COLOR_MAP

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def allianceauth_discordbot_installed() -> bool:
    return apps.is_installed(app_name="aadiscordbot")


def discordnotify_installed() -> bool:
    return apps.is_installed(app_name="discordnotify")


# pylint: disable=import-outside-toplevel, unused-import
def discordproxy_installed() -> bool:
    try:
        # Third Party
        from discordproxy.client import DiscordClient
    except ModuleNotFoundError:
        return False
    return True


# pylint: disable=import-outside-toplevel, import-self, no-name-in-module
def _discordbot_send_direct_message(
    user_id: int,
    title: str,
    message: str,
    embed_message: bool = True,
    level: str = "info",
) -> None:
    """Send a private message to a user via discordbot"""

    # Third Party
    from aadiscordbot.tasks import send_message
    from discord import Embed

    embed = Embed(
        title=str(title),
        description=message,
        color=DISCORD_EMBED_COLOR_MAP.get(level),
        timestamp=timezone.now(),
    )

    if embed_message is True:
        send_message(user_id=user_id, embed=embed)
    else:
        send_message(user_id=user_id, message=f"**{title}**\n\n{message}")


def _discordproxy_send_direct_message(
    user_id: int,
    title: str,
    message: str,
    embed_message: bool = True,
    level: str = "info",
):
    """Send a direct message to a user via discordproxy"""
    # Third Party
    from discordproxy.client import DiscordClient

    client = DiscordClient()
    if embed_message is True:
        # Third Party
        from discordproxy.discord_api_pb2 import Embed

        footer = Embed.Footer(text=str(__title__))
        embed = Embed(
            title=str(title),
            description=message,
            color=DISCORD_EMBED_COLOR_MAP.get(level),
            timestamp=timezone.now().isoformat(),
            footer=footer,
        )

        client.create_direct_message(user_id=user_id, embed=embed)
    else:
        client.create_direct_message(
            user_id=user_id, content=f"**{title}**\n\n{message}"
        )


@shared_task
def send_user_notification(
    user_id: int,
    title: str,
    message: str,
    embed_message: bool = True,
    level: str = "info",
) -> None:
    """Send a notification as a Celery Task to a user via Discord or AllianceAuth notify."""
    try:
        user = User.objects.get(pk=user_id)
    except User.DoesNotExist:
        logger.warning(
            "User with ID %s does not exist. Notification not sent.", user_id
        )
        return

    getattr(notify, level)(user=user, title=title, message=message)

    if hasattr(user, "discord"):  # Check if the user has a Discord account
        if allianceauth_discordbot_installed():
            logger.debug(
                "Discord bot is installed, messages will be sent via Discord bot."
            )
            _discordbot_send_direct_message(
                user_id=int(user.discord.uid),
                title=title,
                message=message,
                embed_message=embed_message,
                level=level,
            )
            return
        if not discordnotify_installed():
            if discordproxy_installed():
                logger.debug(
                    "Discord proxy is installed, messages will be sent via Discord proxy."
                )
                _discordproxy_send_direct_message(
                    user_id=int(user.discord.uid),
                    title=title,
                    message=message,
                    embed_message=embed_message,
                    level=level,
                )
            return
        logger.debug("No message sender found, sending of Discord message canceled.")
    else:
        logger.debug(
            "User %s does not have a Discord account, sending of Discord message canceled.",
            user.username,
        )



# Standard Library
from types import SimpleNamespace
from typing import Any

# Django
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.esi_testing import (
    BravadoOperationStub,
    EsiClientStub,
    EsiEndpoint,
    _EsiMethod,
    parse_datetime,
)


class EsiEndpoint(EsiEndpoint):
    def __init__(self, *args, return_response=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.return_response = return_response


class _EsiMethodOpenApi(_EsiMethod):
    """An ESI method that can be called from the ESI client using OpenAPI client."""

    @staticmethod
    def _convert_values(data) -> Any:
        def convert_dict(item):
            if isinstance(item, dict):
                for key, value in item.items():
                    if isinstance(value, str):
                        try:
                            if parsed_datetime := parse_datetime(value):
                                item[key] = parsed_datetime.replace(tzinfo=timezone.utc)
                        except ValueError:
                            pass

        def dict_to_obj(obj):
            if isinstance(obj, dict):
                return SimpleNamespace(**{k: dict_to_obj(v) for k, v in obj.items()})
            elif isinstance(obj, list):
                return [dict_to_obj(i) for i in obj]
            else:
                return obj

        if isinstance(data, list):
            for row in data:
                convert_dict(row)
        else:
            convert_dict(data)

        return dict_to_obj(data)

    def call(self, **kwargs):
        result_obj = super().call(**kwargs)
        if self._endpoint.return_response:
            return BravadoOperationStub(result_obj.result(), also_return_response=True)
        return result_obj


class EsiClientStubOpenApi(EsiClientStub):
    """ESI Client Stub that uses OpenAPI generated client."""

    def _add_endpoint(self, endpoint: EsiEndpoint):
        if not hasattr(self, endpoint.category):
            setattr(self, endpoint.category, type(endpoint.category, (object,), {}))
        endpoint_category = getattr(self, endpoint.category)
        if not hasattr(endpoint_category, endpoint.method):
            setattr(
                endpoint_category,
                endpoint.method,
                _EsiMethodOpenApi(
                    endpoint=endpoint,
                    testdata=self._testdata,
                    http_error=self._http_error,
                ).call,
            )
        else:
            raise ValueError(f"Endpoint for {endpoint} already defined!")



# Generated by Django 4.2.11 on 2025-02-19 23:15

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("skillfarm", "0002_skillfarmaudit_last_notification_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="characterskill",
            name="name",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="characterskillqueueentry",
            name="has_no_skillqueue",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="characterskillqueueentry",
            name="last_check",
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="characterskillqueueentry",
            name="name",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="skillfarmaudit",
            name="name",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="skillfarmsetup",
            name="name",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name="skillfarmaudit",
            name="id",
            field=models.AutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



# Standard Library
from unittest.mock import patch

# Django
from django.db.utils import Error
from django.test import TestCase, override_settings
from django.utils import timezone

# Alliance Auth
from allianceauth.authentication.models import UserProfile

# AA Skillfarm
from skillfarm import tasks
from skillfarm.models.skillfarmaudit import SkillFarmAudit
from skillfarm.tests.testdata.allianceauth import load_allianceauth
from skillfarm.tests.testdata.eveuniverse import load_eveuniverse
from skillfarm.tests.testdata.skillfarm import (
    add_skillfarmaudit_character_to_user,
    create_evetypeprice,
    create_skill_character,
    create_skillfarm_character,
    create_skillsetup_character,
    create_update_status,
    create_user_from_evecharacter_with_access,
)

TASK_PATH = "skillfarm.tasks"


@patch(TASK_PATH + ".update_character", spec=True)
class TestUpdateAllSkillfarm(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.audit = create_skillfarm_character(1001)

    def test_should_update_all_skillfarm(self, mock_update_all_skillfarm):
        # when
        tasks.update_all_skillfarm()
        # then
        self.assertTrue(mock_update_all_skillfarm.apply_async.called)


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(TASK_PATH + ".chain", spec=True)
@patch(TASK_PATH + ".logger", spec=True)
class TestUpdateCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.audit = create_skillfarm_character(1001)

    def test_update_character_should_no_updated(self, mock_logger, __):
        # when
        tasks.update_character(self.audit.pk)
        # then
        mock_logger.info.assert_called_once_with(
            "No updates needed for %s",
            self.audit.character.character_name,
        )

    def test_update_character_should_update(self, mock_logger, mock_chain):
        # given
        create_update_status(
            self.audit,
            section=SkillFarmAudit.UpdateSection.SKILLS,
            is_success=True,
            error_message="",
            has_token_error=False,
            last_run_at=None,
            last_run_finished_at=None,
            last_update_at=None,
            last_update_finished_at=None,
        )

        # when
        tasks.update_character(self.audit.pk)
        # then
        mock_chain.assert_called_once()


@patch(TASK_PATH + ".SkillFarmAudit.objects.filter", spec=True)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestCheckSkillfarmNotification(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.user2, cls.character_ownership2 = create_user_from_evecharacter_with_access(
            1002
        )
        cls.audit = add_skillfarmaudit_character_to_user(cls.user, 1001)
        cls.audit2 = add_skillfarmaudit_character_to_user(cls.user2, 1002)
        cls.audit3 = add_skillfarmaudit_character_to_user(cls.user2, 1003)

    def _set_notifiaction_status(self, audits, status):
        for audit in audits:
            audit.notification = status
            audit.save()

    def test_no_notification_should_return_false(self, mock_audit_filter):
        audits = [self.audit, self.audit2]
        self._set_notifiaction_status(audits, False)
        mock_audit_filter.return_value = audits
        # when
        tasks.check_skillfarm_notifications()
        # then
        for audit in audits:
            self.assertFalse(audit.notification_sent)
            self.assertIsNone(audit.last_notification)

    def test_notifiaction_with_no_skillsetup_should_return_false(
        self, mock_audit_filter
    ):
        audits = [self.audit, self.audit2, self.audit3]
        self._set_notifiaction_status(audits, True)
        mock_audit_filter.return_value = audits
        # when
        tasks.check_skillfarm_notifications()
        # then
        for audit in audits:
            self.assertFalse(audit.notification_sent)
            self.assertIsNone(audit.last_notification)


@patch(TASK_PATH + ".SkillFarmAudit.objects.filter", spec=True)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestCheckSkillfarmNotificationSuccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.audit = add_skillfarmaudit_character_to_user(cls.user, 1001)
        cls.user2, cls.character_ownership2 = create_user_from_evecharacter_with_access(
            1002
        )
        cls.audit2 = add_skillfarmaudit_character_to_user(cls.user2, 1002)

        cls.skill = create_skill_character(
            character_id=cls.audit.character.character_id,
            evetype_id=1,
            skillpoints=500000,
            trained_level=5,
            active_level=5,
        )
        cls.skillsetup = create_skillsetup_character(
            character_id=cls.audit.character.character_id, skillset=["skill1"]
        )

        cls.skill2 = create_skill_character(
            character_id=cls.audit2.character.character_id,
            evetype_id=2,
            skillpoints=500000,
            trained_level=5,
            active_level=5,
        )
        cls.skillsetup2 = create_skillsetup_character(
            character_id=cls.audit2.character.character_id, skillset=["skill2"]
        )

    def _set_notifiaction_status(self, audits, status):
        for audit in audits:
            audit.notification = status
            audit.save()

    def test_notifiaction_with_skillsetup_should_return_true(self, mock_audit_filter):
        audits = [self.audit, self.audit2]
        self._set_notifiaction_status(audits, True)

        mock_audit_filter.return_value = audits
        # when
        tasks.check_skillfarm_notifications()
        # then
        for audit in audits:
            self.assertTrue(audit.notification_sent)
            self.assertIsNotNone(audit.last_notification)


@patch(TASK_PATH + ".SkillFarmAudit.objects.filter", spec=True)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestCheckSkillfarmNotificationError(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.audit = add_skillfarmaudit_character_to_user(cls.user, 1001)

    def _set_notifiaction_status(self, audits, status):
        for audit in audits:
            audit.notification = status
            audit.save()

    @patch(TASK_PATH + ".logger", spec=True)
    def test_notifiaction_no_main_should_return_false(
        self, mock_logger, mock_audit_filter
    ):
        audits = [self.audit]
        self._set_notifiaction_status(audits, True)

        userprofile = UserProfile.objects.get(user=self.user)
        userprofile.main_character = None
        userprofile.save()
        self.character_ownership.delete()
        self.audit.refresh_from_db()

        mock_audit_filter.return_value = audits
        # when
        tasks.check_skillfarm_notifications()
        # then
        for audit in audits:
            self.assertFalse(audit.notification_sent)
            self.assertIsNone(audit.last_notification)
            mock_logger.warning.assert_called_once_with(
                "Main Character not found for %s, skipping notification",
                self.audit.character.character_name,
            )


@patch(TASK_PATH + ".requests.get", spec=True)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestSkillfarmPrices(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.price = create_evetypeprice(3, buy=100, sell=200, updated_at=timezone.now())
        cls.price2 = create_evetypeprice(
            2, buy=300, sell=400, updated_at=timezone.now()
        )

        cls.json = {
            3: {
                "buy": {"percentile": 100},
                "sell": {"percentile": 200},
            }
        }

    @patch(TASK_PATH + ".EveTypePrice.objects.all", spec=True)
    @patch(TASK_PATH + ".logger", spec=True)
    def test_update_prices_should_update_nothing(
        self, mock_logger, mock_prices, mock_requests
    ):
        mock_prices.return_value = []
        mock_response = mock_requests.return_value
        mock_response.json.return_value = self.json

        # when
        tasks.update_all_prices()
        # then
        mock_logger.info.assert_called_once_with("No Prices to update")

    def test_should_update_prices(self, mock_requests):
        mock_response = mock_requests.return_value
        mock_response.json.return_value = self.json
        # when
        tasks.update_all_prices()
        # then
        self.assertAlmostEqual(self.price.buy, 100)
        self.assertAlmostEqual(self.price.sell, 200)
        self.assertIsNotNone(self.price.updated_at)

    def test_update_prices_should_only_update_existing(self, mock_requests):
        mock_response = mock_requests.return_value
        changed_json = self.json.copy()
        changed_json.update(
            {
                4: {
                    "buy": {"percentile": 300},
                    "sell": {"percentile": 400},
                }
            }
        )
        mock_response.json.return_value = changed_json
        # when
        tasks.update_all_prices()
        # then
        self.assertAlmostEqual(self.price.buy, 100)
        self.assertAlmostEqual(self.price.sell, 200)
        self.assertIsNotNone(self.price.updated_at)

    @patch(TASK_PATH + ".EveTypePrice.objects.bulk_update", spec=True)
    @patch(TASK_PATH + ".logger", spec=True)
    def test_update_prices_should_raise_exception(
        self, mock_logger, mock_bulk_update, mock_requests
    ):
        mock_response = mock_requests.return_value
        mock_response.json.return_value = self.json
        error_instance = Error("Error")
        mock_bulk_update.side_effect = error_instance
        # when
        tasks.update_all_prices()
        # then
        mock_logger.error.assert_called_once_with(
            "Error updating prices: %s", error_instance
        )
